'''
This module contains a high-level API to SNMP functions.

The arguments and return values of these functions have types which are
internal to ``puresnmp`` (subclasses of :py:class:`puresnmp.x690.Type`).

Alternatively, there is :py:mod:`puresnmp.aio.api.pythonic` which converts
these values into pure Python types. This makes day-to-day programming a bit
easier but loses type information which may be useful in some edge-cases. In
such a case it's recommended to use :py:mod:`puresnmp.aio.api.raw`.
'''
from __future__ import unicode_literals

import logging
import sys
from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Tuple, cast, List
from warnings import warn

from ...const import DEFAULT_TIMEOUT, ERRORS_STRICT, ERRORS_WARN, Version
from ...exc import FaultySNMPImplementation, NoSuchOID, SnmpError
from ...pdu import (
    END_OF_MIB_VIEW,
    BulkGetRequest,
    GetNextRequest,
    GetRequest,
    GetResponse,
    SetRequest,
    VarBind
)
from ...typevars import PyType, TWrappedPyType
from ...util import BulkResult  # NOQA (must be here for type detection)
from ...util import get_unfinished_walk_oids, group_varbinds
from ...x690.types import (
    Integer,
    ObjectIdentifier,
    OctetString,
    Sequence,
    Type
)
from ...x690.util import tablify, to_bytes
from ..transport import Transport

if TYPE_CHECKING:  # pragma: no cover
    # pylint: disable=unused-import, invalid-name, ungrouped-imports
    from typing import Type as TType
    from typing import (
        AsyncGenerator,
        Callable,
        Coroutine,
        Dict,
        Set,
        TypeVar,
        Union,
    )
    TWalkResponse = AsyncGenerator[VarBind, None]
    TFetcher = Callable[[str, str, List[str], int, int],
                        Coroutine[Any, Any, List[VarBind]]]

try:
    unicode = unicode  # type: Callable[[Any], str]
except NameError:
    # pylint: disable=invalid-name
    unicode = str  # type: Callable[[Any], str]

_set = set

LOG = logging.getLogger(__name__)
OID = ObjectIdentifier.from_string


async def get(ip, community, oid, port=161, timeout=DEFAULT_TIMEOUT):
    # type: ( str, str, str, int, int ) -> Type[PyType]
    """
    Executes a simple SNMP GET request and returns a pure Python data
    structure.

    Example::

        >>> await get('192.168.1.1', 'private', '1.2.3.4')
        'non-functional example'
    """
    result = await multiget(
        ip, community, [oid], port, timeout=timeout)
    return result[0]


async def multiget(ip, community, oids, port=161, timeout=DEFAULT_TIMEOUT):
    # type: ( str, str, List[str], int, int ) -> List[Type[PyType]]
    """
    Executes an SNMP GET request with multiple OIDs and returns a list of pure
    Python objects. The order of the output items is the same order as the OIDs
    given as arguments.

    Example::

        >>> await multiget('192.168.1.1', 'private', ['1.2.3.4', '1.2.3.5'])
        ['non-functional example', 'second value']
    """
    transport = Transport(timeout=timeout)

    parsed_oids = [OID(oid) for oid in oids]

    packet = Sequence(
        Integer(Version.V2C),
        OctetString(community),
        GetRequest(transport.get_request_id(), *parsed_oids)
    )

    response = await transport.send(ip, port, to_bytes(packet))
    raw_response = cast(
        Tuple[Any, Any, GetResponse],
        Sequence.from_bytes(response)
    )

    output = [value for _, value in raw_response[2].varbinds]
    if len(output) != len(oids):
        raise SnmpError('Unexpected response. Expected %d varbind, '
                        'but got %d!' % (len(oids), len(output)))
    return output


async def getnext(ip, community, oid, port=161, timeout=DEFAULT_TIMEOUT):
    # type: (str, str, str, int, int) -> VarBind
    """
    Executes a single SNMP GETNEXT request (used inside *walk*).

    Example::

        >>> await getnext('192.168.1.1', 'private', '1.2.3')
        VarBind(ObjectIdentifier(1, 2, 3, 0), 'non-functional example')
    """
    result = await multigetnext(
        ip, community, [oid], port, timeout=timeout)
    return result[0]


async def multigetnext(ip, community, oids, port=161, timeout=DEFAULT_TIMEOUT):
    # type: (str, str, List[str], int, int) -> List[VarBind]
    """
    Executes a single multi-oid GETNEXT request.

    The request sends one packet to the remote host requesting the value of the
    OIDs following one or more given OIDs.

    Example::

        >>> await multigetnext('192.168.1.1', 'private', ['1.2.3', '1.2.4'])
        [
            VarBind(ObjectIdentifier(1, 2, 3, 0), 'non-functional example'),
            VarBind(ObjectIdentifier(1, 2, 4, 0), 'second value')
        ]
    """
    transport = Transport(timeout=timeout)
    request = GetNextRequest(transport.get_request_id(), *oids)
    packet = Sequence(
        Integer(Version.V2C),
        OctetString(community),
        request
    )
    response = await transport.send(ip, port, to_bytes(packet))
    raw_response = cast(
        Tuple[Any, Any, GetResponse],
        Sequence.from_bytes(response)
    )
    response_object = raw_response[2]
    if len(response_object.varbinds) != len(oids):
        raise SnmpError(
            'Invalid response! Expected exactly %d varbind, '
            'but got %d' % (len(oids), len(response_object.varbinds)))

    output = []
    for oid, value in response_object.varbinds:
        if value is END_OF_MIB_VIEW:
            break
        output.append(VarBind(oid, value))  # type: ignore

    # Verify that the OIDs we retrieved are successors of the requested OIDs.
    for requested, retrieved in zip(oids, output):
        if not OID(requested) < retrieved.oid:
            stringified = unicode(retrieved.oid)  # TODO remove when Py2 is dropped
            raise FaultySNMPImplementation(
                'The OID %s is not a successor of %s!' %
                (stringified, requested))
    return output


async def walk(ip, community, oid, port=161,
               timeout=DEFAULT_TIMEOUT, errors=ERRORS_STRICT):
    # type: (str, str, str, int, int, str) -> TWalkResponse
    """
    Executes a sequence of SNMP GETNEXT requests and returns a generator over
    :py:class:`~puresnmp.pdu.VarBind` instances.

    The generator stops when hitting an OID which is *not* a sub-node of the
    given start OID or at the end of the tree (whichever comes first).

    Example::

        >>> walk('127.0.0.1', 'private', '1.3.6.1.2.1.1')
        <async_generator object walk at 0x7fa2f775cf68>

        >>> from pprint import pprint
        >>> result = walk('127.0.0.1', 'private', '1.3.6.1.2.1.3')
        >>> res = []
        >>> async for x in gen:
        ...     res.append(x)
        ...
        >>> pprint(res)
        [VarBind(oid=ObjectIdentifier((1, 3, 6, 1, 2, 1, 3, 1, 1, 1, 24, 1, 172, 17, 0, 1)), value=24),
         VarBind(oid=ObjectIdentifier((1, 3, 6, 1, 2, 1, 3, 1, 1, 2, 24, 1, 172, 17, 0, 1)), value=b'\\x02B\\xef\\x14@\\xf5'),
         VarBind(oid=ObjectIdentifier((1, 3, 6, 1, 2, 1, 3, 1, 1, 3, 24, 1, 172, 17, 0, 1)), value=64, b'\\xac\\x11\\x00\\x01')]
    """

    gen = multiwalk(ip, community, [oid], port, timeout=timeout, errors=errors)
    async for varbind in gen:
        yield varbind


async def multiwalk(
        ip, community, oids,
        port=161, timeout=DEFAULT_TIMEOUT, fetcher=multigetnext,
        errors=ERRORS_STRICT):
    # type: (str, str, List[str], int, int, TFetcher, str) -> TWalkResponse
    """
    Executes a sequence of SNMP GETNEXT requests and returns a generator over
    :py:class:`~puresnmp.pdu.VarBind` instances.

    This is the same as :py:func:`~.walk` except that it is capable of
    iterating over multiple OIDs at the same time.

    Example::

        >>> multiwalk('127.0.0.1', 'private', [
        ...     '1.3.6.1.2.1.1', '1.3.6.1.4.1.1'])
        <async_generator object multiwalk at 0x7fa2f775cf68>
    """
    LOG.debug('Walking on %d OIDs using %s', len(oids), fetcher.__name__)

    varbinds = await fetcher(ip, community, oids, port, timeout)
    requested_oids = [OID(oid) for oid in oids]
    grouped_oids = group_varbinds(varbinds, requested_oids)
    unfinished_oids = get_unfinished_walk_oids(grouped_oids)
    LOG.debug('%d of %d OIDs need to be continued',
              len(unfinished_oids),
              len(oids))
    yielded = _set([])  # type: ignore
    for var in sorted(grouped_oids.values()):
        for varbind in var:
            containment = [varbind.oid in _ for _ in requested_oids]
            if not any(containment) or varbind.oid in yielded:
                LOG.debug('Unexpected device response: Returned VarBind %s '
                          'was either not contained in the requested tree or '
                          'appeared more than once. Skipping!', varbind)
                continue
            yielded.add(varbind.oid)
            yield varbind

    # As long as we have unfinished OIDs, we need to continue the walk for
    # those.
    while unfinished_oids:
        next_fetches = [_[1].value.oid for _ in unfinished_oids]
        next_fetches_str = [unicode(_) for _ in next_fetches]
        try:
            varbinds = await fetcher(
                ip, community, next_fetches_str, port, timeout)
        except NoSuchOID:
            # Reached end of OID tree, finish iteration
            break
        except FaultySNMPImplementation as exc:
            if errors == ERRORS_WARN:
                LOG.warning('SNMP walk aborted prematurely due to faulty SNMP '
                            'implementation on device %r! Upon running a '
                            'GetNext on OIDs %r it returned the following '
                            'error: %s', ip, next_fetches_str, exc)
                break
            raise
        grouped_oids = group_varbinds(varbinds,
                                      next_fetches,
                                      user_roots=requested_oids)
        unfinished_oids = get_unfinished_walk_oids(grouped_oids)
        LOG.debug('%d of %d OIDs need to be continued',
                  len(unfinished_oids),
                  len(oids))
        for var in sorted(grouped_oids.values()):
            for varbind in var:
                containment = [varbind.oid in _ for _ in requested_oids]
                if not any(containment) or varbind.oid in yielded:
                    continue
                yielded.add(varbind.oid)
                yield varbind


async def set(ip, community, oid, value, port=161, timeout=DEFAULT_TIMEOUT):  # pylint: disable=redefined-builtin
    # type: (str, str, str, TWrappedPyType, int, int) -> TWrappedPyType
    """
    Executes a simple SNMP SET request. The result is returned as pure Python
    data structure. The value must be a subclass of
    :py:class:`~puresnmp.x690.types.Type`.

    Example::

        >>> await set('127.0.0.1', 'private', '1.3.6.1.2.1.1.4.0',
        ...     OctetString(b'I am contact'))
        b'I am contact'
    """

    result = await multiset(
        ip, community, [(oid, value)], port, timeout=timeout)
    return result[oid.lstrip('.')]


async def multiset(ip, community, mappings, port=161, timeout=DEFAULT_TIMEOUT):
    # type: (str, str, List[Tuple[str, TWrappedPyType]], int, int) -> Dict[str, TWrappedPyType]
    """
    Executes an SNMP SET request on multiple OIDs. The result is returned as
    pure Python data structure.

    Fake Example::

        >>> await multiset(
        ...     '127.0.0.1', 'private',
        ...     [('1.2.3', OctetString(b'foo')),
        ...     ('2.3.4', OctetString(b'bar'))])
        {'1.2.3': b'foo', '2.3.4': b'bar'}
    """
    transport = Transport(timeout=timeout)

    if any([not isinstance(v, Type) for k, v in mappings]):
        raise TypeError('SNMP requires typing information. The value for a '
                        '"set" request must be an instance of "Type"!')

    binds = [VarBind(OID(k), v)
             for k, v in mappings]

    request = SetRequest(transport.get_request_id(), binds)
    packet = Sequence(Integer(Version.V2C),
                      OctetString(community),
                      request)
    response = await transport.send(ip, port, to_bytes(packet))
    raw_response = cast(
        Tuple[Any, Any, GetResponse],
        Sequence.from_bytes(response)
    )
    output = {
        unicode(oid): value
        for oid, value in raw_response[2].varbinds
    }
    if len(output) != len(mappings):
        raise SnmpError('Unexpected response. Expected %d varbinds, '
                        'but got %d!' % (len(mappings), len(output)))
    return output


async def bulkget(
        ip, community, scalar_oids, repeating_oids, max_list_size=1,
        port=161, timeout=DEFAULT_TIMEOUT):
    # type: (str, str, List[str], List[str], int, int, int) -> BulkResult
    """
    Runs a "bulk" get operation and returns a :py:class:`~.BulkResult`
    instance.  This contains both a mapping for the scalar variables (the
    "non-repeaters") and an OrderedDict instance containing the remaining list
    (the "repeaters").

    The OrderedDict is ordered the same way as the SNMP response (whatever the
    remote device returns).

    This operation can retrieve both single/scalar values *and* lists of values
    ("repeating values") in one single request. You can for example retrieve
    the hostname (a scalar value), the list of interfaces (a repeating value)
    and the list of physical entities (another repeating value) in one single
    request.

    Note that this behaves like a **getnext** request for scalar values! So you
    will receive the value of the OID which is *immediately following* the OID
    you specified for both scalar and repeating values!

    :param scalar_oids: contains the OIDs that should be fetched as single
        value.
    :param repeating_oids: contains the OIDs that should be fetched as list.
    :param max_list_size: defines the max length of each list.

    Example::

        >>> ip = '192.168.1.1'
        >>> community = 'private'
        >>> result = await bulkget(
        ...     ip,
        ...     community,
        ...     scalar_oids=['1.3.6.1.2.1.1.1',
        ...                  '1.3.6.1.2.1.1.2'],
        ...     repeating_oids=['1.3.6.1.2.1.3.1',
        ...                     '1.3.6.1.2.1.5.1'],
        ...     max_list_size=10)
        BulkResult(
            scalars={'1.3.6.1.2.1.1.2.0': '1.3.6.1.4.1.8072.3.2.10',
                     '1.3.6.1.2.1.1.1.0': b'Linux aafa4dce0ad4 4.4.0-28-'
                                          b'generic #47-Ubuntu SMP Fri Jun 24 '
                                          b'10:09:13 UTC 2016 x86_64'},
            listing=OrderedDict([
                ('1.3.6.1.2.1.3.1.1.1.10.1.172.17.0.1', 10),
                ('1.3.6.1.2.1.5.1.0', b'\x01'),
                ('1.3.6.1.2.1.3.1.1.2.10.1.172.17.0.1', b'\x02B\x8e>\x9ee'),
                ('1.3.6.1.2.1.5.2.0', b'\x00'),
                ('1.3.6.1.2.1.3.1.1.3.10.1.172.17.0.1', b'\xac\x11\x00\x01'),
                ('1.3.6.1.2.1.5.3.0', b'\x00'),
                ('1.3.6.1.2.1.4.1.0', 1),
                ('1.3.6.1.2.1.5.4.0', b'\x01'),
                ('1.3.6.1.2.1.4.3.0', b'\x00\xb1'),
                ('1.3.6.1.2.1.5.5.0', b'\x00'),
                ('1.3.6.1.2.1.4.4.0', b'\x00'),
                ('1.3.6.1.2.1.5.6.0', b'\x00'),
                ('1.3.6.1.2.1.4.5.0', b'\x00'),
                ('1.3.6.1.2.1.5.7.0', b'\x00'),
                ('1.3.6.1.2.1.4.6.0', b'\x00'),
                ('1.3.6.1.2.1.5.8.0', b'\x00'),
                ('1.3.6.1.2.1.4.7.0', b'\x00'),
                ('1.3.6.1.2.1.5.9.0', b'\x00'),
                ('1.3.6.1.2.1.4.8.0', b'\x00'),
                ('1.3.6.1.2.1.5.10.0', b'\x00')]))
    """
    transport = Transport(timeout=timeout)

    scalar_oids = scalar_oids or []  # protect against empty values
    repeating_oids = repeating_oids or []  # protect against empty values

    oids = [
        OID(oid) for oid in scalar_oids
    ] + [
        OID(oid) for oid in repeating_oids
    ]

    non_repeaters = len(scalar_oids)

    packet = Sequence(
        Integer(Version.V2C),
        OctetString(community),
        BulkGetRequest(
            transport.get_request_id(), non_repeaters, max_list_size, *oids)
    )

    response = await transport.send(ip, port, to_bytes(packet))
    raw_response = cast(
        Tuple[Any, Any, GetResponse],
        Sequence.from_bytes(response)
    )

    # See RFC=3416 for details of the following calculation
    n = min(non_repeaters, len(oids))
    m = max_list_size
    r = max(len(oids) - n, 0)  # pylint: disable=invalid-name
    expected_max_varbinds = n + (m * r)

    _, _, get_response = raw_response
    n_retrieved_varbinds = len(get_response.varbinds)
    if n_retrieved_varbinds > expected_max_varbinds:
        raise SnmpError('Unexpected response. Expected no more than %d '
                        'varbinds, but got %d!' % (
                            expected_max_varbinds, n_retrieved_varbinds))

    # cut off the scalar OIDs from the listing(s)
    scalar_tmp = get_response.varbinds[0:len(scalar_oids)]
    repeating_tmp = get_response.varbinds[len(scalar_oids):]

    # prepare output for scalar OIDs
    scalar_out = {
        unicode(oid): value
        for oid, value in scalar_tmp
    }

    # prepare output for listing
    repeating_out = OrderedDict()  # type: Dict[str, Type]
    for oid, value in repeating_tmp:
        if value is END_OF_MIB_VIEW:
            break
        repeating_out[unicode(oid)] = value

    return BulkResult(scalar_out, repeating_out)


def _bulkwalk_fetcher(bulk_size=10):
    # type: (int) -> Callable[[str, str, List[str], int, int], Coroutine[ Any, Any, List[VarBind]]]
    """
    Create a bulk fetcher with a fixed limit on "repeatable" OIDs.
    """

    async def fetcher(ip, community, oids, port=161, timeout=DEFAULT_TIMEOUT):
        # type: (str, str, List[str], int, int) -> List[VarBind]
        '''
        Executes a SNMP BulkGet request.
        '''
        result = await bulkget(
            ip, community, [], oids,
            max_list_size=bulk_size,
            port=port, timeout=timeout)
        return [VarBind(OID(k), v)
                for k, v in result.listing.items()]

    if sys.version_info < (3, 0):
        fetcher.__name__ = str('_bulkwalk_fetcher(%d)' % bulk_size)
    else:
        fetcher.__name__ = '_bulkwalk_fetcher(%d)' % bulk_size
    return fetcher


async def bulkwalk(ip, community, oids, bulk_size=10, port=161,
                   timeout=DEFAULT_TIMEOUT):
    # type: (str, str, List[str], int, int, int) -> TWalkResponse
    """
    More efficient implementation of :py:func:`~.walk`. It uses
    :py:func:`~.bulkget` under the hood instead of :py:func:`~.getnext`.

    Just like :py:func:`~.multiwalk`, it returns a generator over
    :py:class:`~puresnmp.pdu.VarBind` instances.

    :param ip: The IP address of the target host.
    :param community: The community string for the SNMP connection.
    :param oids: A list of base OIDs to use in the walk operation.
    :param bulk_size: How many varbinds to request from the remote host with
        one request.
    :param port: The TCP port of the remote host.
    :param timeout: The TCP timeout for network calls

    Example::

        >>> from puresnmp.api.aio_raw import bulkwalk
        >>> ip = '127.0.0.1'
        >>> community = 'private'
        >>> oids = [
        ...     '1.3.6.1.2.1.2.2.1.2',   # name
        ...     '1.3.6.1.2.1.2.2.1.6',   # MAC
        ...     '1.3.6.1.2.1.2.2.1.22',  # ?
        ... ]
        >>> result = bulkwalk(ip, community, oids)
        >>> async for row in result:
        ...     print(row)
        VarBind(oid=ObjectIdentifier((1, 3, 6, 1, 2, 1, 2, 2, 1, 2, 1)), value=b'lo')
        VarBind(oid=ObjectIdentifier((1, 3, 6, 1, 2, 1, 2, 2, 1, 6, 1)), value=b'')
        VarBind(oid=ObjectIdentifier((1, 3, 6, 1, 2, 1, 2, 2, 1, 22, 1)), value='0.0')
        VarBind(oid=ObjectIdentifier((1, 3, 6, 1, 2, 1, 2, 2, 1, 2, 38)), value=b'eth0')
        VarBind(oid=ObjectIdentifier((1, 3, 6, 1, 2, 1, 2, 2, 1, 6, 38)), value=b'\x02B\xac\x11\x00\x02')
        VarBind(oid=ObjectIdentifier((1, 3, 6, 1, 2, 1, 2, 2, 1, 22, 38)), value='0.0')
    """

    if not isinstance(oids, list):
        raise TypeError('OIDS need to be passed as list!')

    result = multiwalk(ip, community, oids, port=port,
                       fetcher=_bulkwalk_fetcher(bulk_size),
                       timeout=timeout)
    async for oid, value in result:
        yield VarBind(oid, value)  # type: ignore


async def table(ip, community, oid, port=161, num_base_nodes=0):
    # type: (str, str, str, int, int) -> List[Dict[str, Any]]
    """
    Fetch an SNMP table

    The resulting output will be a list of dictionaries where each dictionary
    corresponds to a row of the table.

    The index of the row will be contained in key ``'0'`` as a string
    representing an OID. This key ``'0'`` is automatically injected by
    ``puresnmp``. Table rows may or may not contain the row-index in other
    columns. This depends on the requested table.

    Each column ID is available as *string*.

    Example output (using fake data):

    >>> table('192.0.2.1', 'private', '1.3.6.1.2.1.2.2')
    [{'0': '1', '1': Integer(1), '2': Counter(30)},
     {'0': '2', '1': Integer(2), '2': Counter(123)}]
    """
    tmp = []  # type: List[Tuple[str, Type[Any]]]
    if num_base_nodes:
        warn('Usage of "num_base_nodes" in table operations is no longer '
             'required', DeprecationWarning)
    else:
        parsed_oid = OID(oid)
        num_base_nodes = len(parsed_oid) + 1  # type: ignore
    varbinds = walk(ip, community, oid, port=port)
    async for varbind in varbinds:
        tmp.append(varbind)  # type: ignore
    as_table = tablify(tmp, num_base_nodes=num_base_nodes)
    return as_table


async def bulktable(ip, community, oid, port=161, num_base_nodes=0, bulk_size=10):
    # type: (str, str, str, int, int, int) -> List[Dict[str, Any]]
    """
    Fetch an SNMP table using "bulk" requests.

    See :py:func:`.table` for more information of the returned structure.

    .. versionadded: 1.7.0
    """
    tmp_raw = []
    if num_base_nodes:
        warn('Usage of "num_base_nodes" in table operations is no longer '
             'required', DeprecationWarning)
    else:
        parsed_oid = OID(oid)
        num_base_nodes = len(parsed_oid) + 1  # type: ignore
    varbinds = bulkwalk(ip, community, [oid], port=port, bulk_size=bulk_size)
    async for varbind in varbinds:
        tmp_raw.append(varbind)
    tmp = cast(List[Tuple[Any, Any]], tmp_raw)
    as_table = tablify(tmp, num_base_nodes=num_base_nodes)
    return as_table



"""
Unit tests for types specified in RFC-2578
"""

import pytest
from puresnmp import types as t


@pytest.mark.parametrize('value, expected', [
    (-42, 0),  # Underflow below threshold
    (-1, 0),  # Underflow at threshold
    (0, 0),  # The minimum value
    (42, 42),  # A normal value
    (2**32-1, 2**32-1),  # max value
    (2**32, 0),  # overflow at threshold
    (2**32+42, 42),  # overflow above threshold
    ((2**32)*2+42, 42),  # overflow above threshold
])
def test_counter(value, expected):
    """
    A counter instance should be a non-negative integer
    """
    instance = t.Counter(value)
    assert instance.value == expected


def test_counter_issue_75():
    """
    GitHub issue #75 reports incorrect counter decoding.

    This test covers this issue.
    """
    data = b'\x41\x04\x84\x43\x20\xf8'
    result = t.Counter.from_bytes(data)
    expected = t.Counter(2218991864)
    assert result == expected


@pytest.mark.parametrize('value, expected', [
    (-42, 0),  # Underflow below threshold
    (-1, 0),  # Underflow at threshold
    (0, 0),  # The minimum value
    (42, 42),  # A normal value
    (2**64-1, 2**64-1),  # max value
    (2**64, 0),  # overflow at threshold
    ((2**64)*2+42, 42),  # overflow above threshold
])
def test_counter64(value, expected):
    """
    A counter instance should be a non-negative integer
    """
    instance = t.Counter64(value)
    assert instance.value == expected



"""
This module contains variour type aliases for type checking
"""
from datetime import datetime, timedelta
from typing import NamedTuple, TypeVar, Union

#: This type represents the data type of an SNMP value in the Python worldj.
PyType = Union[str, bytes, int, datetime, timedelta, None, float]


#: This is a type variable helps to properly hint functions which take in a
#: "puresnmp" type and return a pythonic type
TWrappedPyType = TypeVar('TWrappedPyType', bound=PyType)


SocketInfo = NamedTuple("SocketInfo", [
    ("address", str),
    ("port", int),
])
SocketResponse = NamedTuple("SocketResponse", [
    ("data", bytes),
    ("info", SocketInfo),
])



"""
Low-Level network transport for asyncio.

This module mainly exist to enable a "seam" for mocking/patching out during
testing.

The module is excluded from coverage. It contains all the "dirty" stuff that's
hard to test.
"""

# TODO (beginner, no-dev): Ignore this file from coverage without adding
#                          "pragma: no cover" to each function.

import asyncio
import logging
from asyncio.events import AbstractEventLoop
from typing import Optional, Tuple, Union

from ..exc import Timeout
from ..transport import Transport as SyncTransport
from ..x690.util import visible_octets

LOG = logging.getLogger(__name__)


class SNMPClientProtocol(asyncio.DatagramProtocol):
    """
    An SNMP Client Protocol suitable for use with create_datagram_endpoint()
    that provide a method to convert the callback based API into a coroutine
    based API.

    """

    def __init__(self, packet, loop):
        # type: (bytes, AbstractEventLoop) -> None
        self.packet = packet
        self.transport = None  # type: Optional[asyncio.DatagramTransport]
        self.loop = loop
        self.future = loop.create_future()

    def connection_made(self, transport):  # type: ignore
        # type: (asyncio.DatagramTransport) -> None
        """
        Sends the SNMP request packet when a connection is made.
        """
        self.transport = transport

        if LOG.isEnabledFor(logging.DEBUG):
            hexdump = visible_octets(self.packet)
            ip, port = self.transport.get_extra_info('peername', ('', ''))
            LOG.debug('Sending packet to %s:%s\n%s', ip, port, hexdump)

        self.transport.sendto(self.packet)

    def connection_lost(self, exc):
        # type: (Optional[Exception]) -> None
        """
        Handles the socket being closed optionally passing on an exception.
        """
        if LOG.isEnabledFor(logging.DEBUG):
            if exc is None:
                LOG.debug('Socket closed')
            else:
                LOG.debug('Connection lost: %s', exc)

        if exc is not None:
            self.future.set_exception(exc)

    def datagram_received(self, data, addr):
        # type: (Union[bytes, str], Tuple[str, int]) -> None
        """
        Receive the data and close the connection.
        """
        if LOG.isEnabledFor(logging.DEBUG) and isinstance(data, bytes):
            hexdump = visible_octets(data)
            LOG.debug('Received packet:\n%s', hexdump)

        self.future.set_result(data)
        if self.transport:
            self.transport.close()

    def error_received(self, exc):
        # type: (Exception) -> None
        """
        Pass the exception along if there is an error.
        """
        if LOG.isEnabledFor(logging.DEBUG):
            LOG.debug('Error received: %s', exc)

        self.future.set_exception(exc)

    async def get_data(self, timeout):
        # type: (int) -> bytes
        """
        Retrieve the response data back into the calling coroutine.
        """
        try:
            return await asyncio.wait_for(self.future, timeout, loop=self.loop)
        except asyncio.TimeoutError:
            if self.transport:
                self.transport.abort()
            raise Timeout("{} second timeout exceeded".format(timeout))


class Transport(SyncTransport):

    async def send(  # type: ignore
            self, ip, port, packet, timeout=6, loop=None):  # pragma: no cover
        # type: ( str, int, bytes, int, Optional[AbstractEventLoop] ) -> bytes
        """
        A coroutine that opens a UDP socket to *ip:port*, sends a packet with
        *bytes* and returns the raw bytes as returned from the remote host.

        If the connection fails due to a timeout, a Timeout exception is
        raised.
        """
        if loop is None:
            loop = asyncio.get_event_loop()

        # family could be specified here (and is in the sync implementation),
        # is it needed? are retries necessary for async implementation?
        transport, protocol = await loop.create_datagram_endpoint(
            lambda: SNMPClientProtocol(packet, loop),  # type: ignore
            remote_addr=(ip, port)
        )

        response = await protocol.get_data(timeout)  # type: ignore

        return response  # type: ignore



# pylint: skip-file

"""
Test the "external" interface.

The "external" interface is what the user sees. It should be pythonic and easy
to use.
"""

from __future__ import unicode_literals

import sys
from datetime import timedelta

import pytest

from puresnmp.aio.api.pythonic import (
    bulkget,
    bulktable,
    bulkwalk,
    get,
    getnext,
    multiget,
    multiset,
    multiwalk,
    set,
    table,
    walk
)
from puresnmp.const import Version
from puresnmp.exc import NoSuchOID, SnmpError
from puresnmp.pdu import BulkGetRequest, GetNextRequest, GetRequest, VarBind
from puresnmp.types import Counter, Gauge, IpAddress
from puresnmp.util import BulkResult
from puresnmp.x690.types import (
    Integer,
    ObjectIdentifier,
    OctetString,
    Sequence,
    to_bytes
)

from .asyncmock import AsyncGenMock, AsyncMock

try:
    from unittest.mock import patch, call
except ImportError:
    from mock import patch, call  # pip install mock


pytestmark = pytest.mark.skipif(sys.version_info < (3, 5),
                                reason="requires python3.5")


class TestGet(object):

    @pytest.mark.asyncio
    async def test_get_string(self):
        expected = (b'Linux d24cf7f36138 4.4.0-28-generic #47-Ubuntu SMP '
                    b'Fri Jun 24 10:09:13 UTC 2016 x86_64')
        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncMock) as mck:
            mck.get.return_value = OctetString(
                b'Linux d24cf7f36138 4.4.0-28-generic #47-Ubuntu SMP '
                b'Fri Jun 24 10:09:13 UTC 2016 x86_64')
            result = await get('::1', 'private', '1.2.3')
        assert result == expected

    @pytest.mark.asyncio
    async def test_get_oid(self):
        expected = ('1.3.6.1.4.1.8072.3.2.10')
        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncMock) as mck:
            mck.get.return_value = ObjectIdentifier.from_string(
                '1.3.6.1.4.1.8072.3.2.10')
            result = await get('::1', 'private', '1.2.3')
        assert result == expected


class TestSet(object):

    @pytest.mark.asyncio
    async def test_set_string(self):
        expected = (b'foo')
        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncMock) as mck:
            mck.multiset.return_value = {
                ObjectIdentifier.from_string('1.2.3'): OctetString(b'foo')
            }
            result = await set('::1', 'private', '1.2.3', OctetString(b'foo'))
        assert result == expected

    @pytest.mark.asyncio
    async def test_set_string_absolute(self):
        expected = (b'foo')
        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncMock) as mck:
            mck.multiset.return_value = {
                ObjectIdentifier.from_string('1.2.3'): OctetString(b'foo')
            }
            result = await set('::1', 'private', '.1.2.3', OctetString(b'foo'))
        assert result == expected


class TestWalk(object):

    @pytest.mark.asyncio
    @pytest.mark.skipif(sys.version_info < (3, 6),
                        reason="requires python3.6")
    async def test_walk(self):
        expected = [VarBind(
            '1.3.6.1.2.1.2.2.1.5.1', 10000000
        ), VarBind(
            '1.3.6.1.2.1.2.2.1.5.13', 4294967295
        )]

        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncGenMock) as mck:
            mck.walk.return_value = [
                VarBind(
                    ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.5.1'),
                    Gauge(10000000)
                ), VarBind(
                    ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.5.13'),
                    Integer(4294967295)
                )]
            result = []
            async for x in walk('::1', 'public', '1.3.6.1.2.1.2.2.1.5'):
                result.append(x)
        assert result == expected


class TestMultiGet(object):

    @pytest.mark.asyncio
    async def test_multiget(self):
        expected = ['1.3.6.1.4.1.8072.3.2.10',
                    b"Linux 7fbf2f0c363d 4.4.0-28-generic #47-Ubuntu SMP Fri "
                    b"Jun 24 10:09:13 UTC 2016 x86_64"]
        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncMock) as mck:
            mck.multiget.return_value = [
                ObjectIdentifier.from_string('1.3.6.1.4.1.8072.3.2.10'),
                OctetString(b"Linux 7fbf2f0c363d 4.4.0-28-generic "
                            b"#47-Ubuntu SMP Fri Jun 24 10:09:13 "
                            b"UTC 2016 x86_64")
            ]
            result = await multiget('::1', 'private', [
                '1.3.6.1.2.1.1.2.0',
                '1.3.6.1.2.1.1.1.0',
            ])
        assert result == expected


class TestMultiWalk(object):

    @pytest.mark.asyncio
    @pytest.mark.skipif(sys.version_info < (3, 6),
                        reason="requires python3.6")
    async def test_multi_walk(self):
        expected = [
            VarBind('1.3.6.1.2.1.2.2.1.1.1', 1),
            VarBind('1.3.6.1.2.1.2.2.1.2.1', b'lo'),
            VarBind('1.3.6.1.2.1.2.2.1.1.78', 78),
            VarBind('1.3.6.1.2.1.2.2.1.2.78', b'eth0')
        ]

        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncGenMock) as mck:
            mck.multiwalk.return_value = [VarBind(
                ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.1.1'),
                1
            ), VarBind(
                ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.2.1'),
                b'lo'
            ), VarBind(
                ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.1.78'),
                78
            ), VarBind(
                ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.2.78'),
                b'eth0'
            )]
            result = []
            async for x in multiwalk('::1', 'public', [
                '1.3.6.1.2.1.2.2.1.1',
                '1.3.6.1.2.1.2.2.1.2'
            ]):
                result.append(x)
        # TODO (advanced): should order matter in the following result?
        assert len(result) == len(expected)


class TestMultiSet(object):

    @pytest.mark.asyncio
    async def test_multiset(self):
        """
        Test setting multiple OIDs at once.

        NOTE: The OID '1.3.6.1.2.1.1.5.0' below is manually edited for
              unit-testing. It probably has a different type in the real world!
        """
        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncMock) as mck:
            mck.multiset.return_value = {
                '1.3.6.1.2.1.1.4.0': OctetString(b'hello@world.com'),
                '1.3.6.1.2.1.1.5.0': OctetString(b'hello@world.com'),
            }
            result = await multiset('::1', 'private', [
                ('1.3.6.1.2.1.1.4.0', OctetString(b'hello@world.com')),
                ('1.3.6.1.2.1.1.5.0', OctetString(b'hello@world.com')),
            ])
        expected = {
            '1.3.6.1.2.1.1.4.0': b'hello@world.com',
            '1.3.6.1.2.1.1.5.0': b'hello@world.com',
        }
        assert result == expected

    @pytest.mark.asyncio
    async def test_multiset_absolute(self):
        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncMock) as mck:
            mck.multiset.return_value = {
                '1.3.6.1.2.1.1.4.0': OctetString(b'hello@world.com'),
                '1.3.6.1.2.1.1.5.0': OctetString(b'hello@world.com'),
            }
            result = await multiset('::1', 'private', [
                ('.1.3.6.1.2.1.1.4.0', OctetString(b'hello@world.com')),
                ('.1.3.6.1.2.1.1.5.0', OctetString(b'hello@world.com')),
            ])
        expected = {
            '1.3.6.1.2.1.1.4.0': b'hello@world.com',
            '1.3.6.1.2.1.1.5.0': b'hello@world.com',
        }
        assert result == expected


class TestGetNext(object):

    @pytest.mark.asyncio
    async def test_getnext(self):
        expected = VarBind('1.3.6.1.6.3.1.1.6.1.0', 354522558)

        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncMock) as mck:
            mck.multigetnext.return_value = [
                VarBind('1.3.6.1.6.3.1.1.6.1.0', Integer(354522558))
            ]
            result = await getnext('::1', 'private', '1.3.6.1.5')
        assert result == expected


class TestGetBulkGet(object):

    @pytest.mark.asyncio
    async def test_bulkget(self):
        expected = BulkResult(
            {'1.3.6.1.2.1.1.1.0': b'Linux 7e68e60fe303 4.4.0-28-generic '
             b'#47-Ubuntu SMP Fri Jun 24 10:09:13 UTC 2016 x86_64'},
            {'1.3.6.1.2.1.3.1.1.1.10.1.172.17.0.1': 10,
             '1.3.6.1.2.1.3.1.1.2.10.1.172.17.0.1': b'\x02B\xe2\xc5\x8d\t',
             '1.3.6.1.2.1.3.1.1.3.10.1.172.17.0.1': b'\xac\x11\x00\x01',
             '1.3.6.1.2.1.4.1.0': 1,
             '1.3.6.1.2.1.4.3.0': 57})

        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncMock) as mck:
            mck.bulkget.return_value = BulkResult({
                '1.3.6.1.2.1.1.1.0': OctetString(
                    b'Linux 7e68e60fe303 4.4.0-28-generic '
                    b'#47-Ubuntu SMP Fri Jun 24 10:09:13 UTC 2016 x86_64')
            }, {
                '1.3.6.1.2.1.3.1.1.1.10.1.172.17.0.1': Integer(10),
                '1.3.6.1.2.1.3.1.1.2.10.1.172.17.0.1': OctetString(
                    b'\x02B\xe2\xc5\x8d\t'),
                '1.3.6.1.2.1.3.1.1.3.10.1.172.17.0.1': IpAddress(
                    b'\xac\x11\x00\x01'),
                '1.3.6.1.2.1.4.1.0': Integer(1),
                '1.3.6.1.2.1.4.3.0': Counter(57)
            })
            result = await bulkget('::1', 'public',
                                   ['1.3.6.1.2.1.1.1'],
                                   ['1.3.6.1.2.1.3.1'],
                                   max_list_size=5)
        assert result == expected


class TestGetBulkWalk(object):

    @pytest.mark.asyncio
    @pytest.mark.skipif(sys.version_info < (3, 6),
                        reason="requires python3.6")
    async def test_bulkwalk(self):
        request_ids = [1001613222, 1001613223, 1001613224]
        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncGenMock) as mck:
            mck.multiwalk.return_value = [
                VarBind('1.3.6.1.2.1.2.2.1.1.1', Integer(1)),
                VarBind('1.3.6.1.2.1.2.2.1.1.10', Integer(10)),
                VarBind('1.3.6.1.2.1.2.2.1.2.1', OctetString(b"lo")),
                VarBind('1.3.6.1.2.1.2.2.1.22.10', ObjectIdentifier(0, 0))
            ]

            result = []
            async for x in bulkwalk('127.0.0.1', 'private', ['1.3.6.1.2.1.2.2'],
                                    bulk_size=20):
                result.append(x)

        expected = [
            VarBind('1.3.6.1.2.1.2.2.1.1.1', 1),
            VarBind('1.3.6.1.2.1.2.2.1.1.10', 10),
            VarBind('1.3.6.1.2.1.2.2.1.2.1', b"lo"),
            VarBind('1.3.6.1.2.1.2.2.1.22.10', '0.0'),
        ]

        assert result == expected


class TestTable(object):

    @pytest.mark.asyncio
    @pytest.mark.skipif(sys.version_info < (3, 6),
                        reason="requires python3.6")
    async def test_table(self):
        with patch('puresnmp.aio.api.pythonic.raw', new_callable=AsyncGenMock) as mck:
            oid = ObjectIdentifier.from_string
            mck.table.return_value = [
                {'0': '1', '1': Integer(1)},
                {'0': '2', '1': Integer(2)},
            ]
            aio_result = table('1.2.3.4', 'private', '1.2')
            result = []
            async for row in aio_result:
                result.append(row)
        expected = [
            {'0': '1', '1': 1},
            {'0': '2', '1': 2},
        ]
        assert len(result) == len(expected)


class TestBulkTable(object):

    @pytest.mark.asyncio
    @pytest.mark.skipif(sys.version_info < (3, 6),
                        reason="requires python3.6")
    async def test_table(self):
        with patch('puresnmp.aio.api.pythonic.raw') as mck:
            oid = ObjectIdentifier.from_string
            mck.bulktable.return_value = [
                {'0': '1', '1': Integer(1)},
                {'0': '2', '1': Integer(2)},
            ]
            aio_result = bulktable('1.2.3.4', 'private', '1.2')
            result = []
            async for row in aio_result:
                result.append(row)
        expected = [
            {'0': '1', '1': 1},
            {'0': '2', '1': 2},
        ]
        assert result == expected



"""
This module contains the high-level functions to access the library.

Care is taken to make this as pythonic as possible and hide as many of the gory
implementations as possible.

This module provides "syntactic sugar" around the lower-level, but almost
identical, module :py:mod:`puresnmp.api.raw`.

The "raw" module returns the variable types unmodified which are all subclasses
of :py:class:`puresnmp.x690.types.Type`.
"""

# TODO (advanced): This module should not make use of it's own functions. The
#     is beginning to be too "thick", containing too much business logic for a
#     mere abstraction layer.
#     module exists as an abstraction layer only. If one function uses a
#     "siblng" function, valuable information is lost. In general, this module


from __future__ import unicode_literals

import logging
from collections import OrderedDict
from datetime import timedelta
from typing import TYPE_CHECKING, TypeVar
from warnings import warn

from ..const import DEFAULT_TIMEOUT, Version
from ..pdu import Trap, VarBind
from ..util import BulkResult
from ..x690.types import ObjectIdentifier, Type
from . import raw

if TYPE_CHECKING:  # pragma: no cover
    # pylint: disable=unused-import, invalid-name
    from typing import Any, Callable, Dict, Generator, List, Tuple, Union
    from puresnmp.typevars import PyType
    TWalkResponse = Generator[VarBind, None, None]
    TFetcher = Callable[[str, str, List[str], int, int],
                        List[VarBind]]
    T = TypeVar('T', bound=PyType)

try:
    unicode = unicode  # type: Callable[[Any], str]
except NameError:
    # pylint: disable=invalid-name
    unicode = str  # type: Callable[[Any], str]

_set = set
LOG = logging.getLogger(__name__)
OID = ObjectIdentifier.from_string


class TrapInfo:
    """
    This class wraps a :py:class:`puresnmp.pdu.Trap` instance (accessible via
    ``raw_trap`` and makes values available as "pythonic" values.
    """

    #: The raw Trap PDU.
    #:
    #: .. warning::
    #:     This will leak data-types which are used internally by ``puresnmp``
    #:     and may change even in minor updates. You should, if possible use
    #:     the values from the properties on this object.  This exist mainly to
    #:     expose values which can be helpful in debugging.  If something is
    #:     missing from the properties, please open a corresponding support
    #:     ticket!
    raw_trap = None

    def __init__(self, raw_trap):
        # type: (Trap) -> None
        self.raw_trap = raw_trap

    def __repr__(self):
        # type: () -> str
        return "<TrapInfo from %s on %s with %d values>" % (
            self.origin,
            self.oid,
            len(self.values),
        )

    @property
    def origin(self):
        # type: () -> str
        """
        Accesses the IP-Address from which the trap was sent

        May be the empty string if the source is unknown
        """
        if self.raw_trap is None or self.raw_trap.source is None:
            return ''
        return self.raw_trap.source.address

    @property
    def uptime(self):
        # type: () -> timedelta
        """
        Returns the uptime of the device.
        """
        return self.raw_trap.varbinds[0].value.pythonize()  # type: ignore

    @property
    def oid(self):
        # type: () -> str
        """
        Returns the Trap-OID
        """
        return self.raw_trap.varbinds[1].value.pythonize()  # type: ignore

    @property
    def values(self):
        # type: () -> Dict[str, Any]
        """
        Returns all the values contained in this trap as dictionary mapping
        OIDs to values.
        """
        output = {}
        for oid_raw, value_raw in self.raw_trap.varbinds[2:]:  # type: ignore
            oid = oid_raw.pythonize()  # type: ignore
            value = value_raw.pythonize()  # type: ignore
            output[oid] = value
        return output


def get(ip, community, oid, port=161, timeout=2, version=Version.V2C):
    # type: (str, str, str, int, int) -> PyType
    """
    Delegates to :py:func:`~puresnmp.api.raw.get` but returns simple Python
    types.

    See the "raw" equivalent for detailed documentation & examples.
    """
    raw_value = raw.get(ip, community, oid, port, timeout=timeout, version=version)
    return raw_value.pythonize()  # type: ignore


def multiget(ip, community, oids, port=161, timeout=DEFAULT_TIMEOUT, version=Version.V2C):
    # type: (str, str, List[str], int, int, int) -> List[PyType]
    """
    Delegates to :py:func:`~puresnmp.api.raw.multiget` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """
    raw_output = raw.multiget(ip, community, oids, port, timeout, version=version)
    pythonized = [value.pythonize() for value in raw_output]
    return pythonized


def getnext(ip, community, oid, port=161, timeout=DEFAULT_TIMEOUT, version=Version.V2C):
    # type: (str, str, str, int, int) -> VarBind
    """
    Delegates to :py:func:`~puresnmp.api.raw.getnext` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """
    result = multigetnext(ip, community, [oid], port, timeout=timeout, version=version)
    return result[0]


def multigetnext(ip, community, oids, port=161, timeout=DEFAULT_TIMEOUT, version=Version.V2C):
    # type: (str, str, List[str], int, int, int) -> List[VarBind]
    """
    Delegates to :py:func:`~puresnmp.api.raw.multigetnext` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """
    raw_output = raw.multigetnext(ip, community, oids, port, timeout, version=version)
    pythonized = [VarBind(oid, value.pythonize())  # type: ignore
                  for oid, value in raw_output]
    return pythonized


def walk(ip, community, oid, port=161, timeout=DEFAULT_TIMEOUT, version=Version.V2C):
    # type: (str, str, str, int, int) -> TWalkResponse
    """
    Delegates to :py:func:`~puresnmp.api.raw.walk` but returns simple Python
    types.

    See the "raw" equivalent for detailed documentation & examples.
    """

    raw_result = raw.walk(ip, community, oid, port, timeout, version=version)
    for raw_oid, raw_value in raw_result:
        yield VarBind(raw_oid, raw_value.pythonize())  # type: ignore


def multiwalk(ip, community, oids, port=161, timeout=DEFAULT_TIMEOUT,
              fetcher=multigetnext, version=Version.V2C):
    # type: (str, str, List[str], int, int, TFetcher, int) -> TWalkResponse
    """
    Delegates to :py:func:`~puresnmp.api.raw.multiwalk` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """
    raw_output = raw.multiwalk(ip, community, oids, port, timeout, fetcher,
                               version=version)
    for oid, value in raw_output:
        if isinstance(value, Type):
            value = value.pythonize()
        yield VarBind(oid, value)  # type: ignore


def set(ip, community, oid, value, port=161, timeout=DEFAULT_TIMEOUT,
        version=Version.V2C):  # pylint: disable=redefined-builtin
    # type: (str, str, str, Type[T], int, int) -> T
    """
    Delegates to :py:func:`~puresnmp.api.raw.set` but returns simple Python
    types.

    See the "raw" equivalent for detailed documentation & examples.
    """

    result = multiset(ip, community, [(oid, value)],  # type: ignore
                      port, timeout=timeout, version=version)
    return result[oid.lstrip('.')]  # type: ignore


def multiset(ip, community, mappings, port=161, timeout=DEFAULT_TIMEOUT,
             version=Version.V2C):
    # type: (str, str, List[Tuple[str, raw.T]], int, int) -> Dict[str, PyType]
    """
    Delegates to :py:func:`~puresnmp.api.raw.multiset` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """

    raw_output = raw.multiset(ip, community, mappings, port, timeout,
                              version=version)
    pythonized = {unicode(oid): value.pythonize()
                  for oid, value in raw_output.items()}
    return pythonized


def bulkget(ip, community, scalar_oids, repeating_oids, max_list_size=1,
            port=161, timeout=DEFAULT_TIMEOUT, version=Version.V2C):
    # type: (str, str, List[str], List[str], int, int, int, int) -> BulkResult
    """
    Delegates to :py:func:`~puresnmp.api.raw.mulkget` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """

    raw_output = raw.bulkget(ip, community, scalar_oids, repeating_oids,
                             max_list_size=max_list_size,
                             port=port,
                             timeout=timeout,
                             version=version)
    pythonized_scalars = {oid: value.pythonize()
                          for oid, value in raw_output.scalars.items()}
    pythonized_list = OrderedDict(
        [(oid, value.pythonize())
         for oid, value in raw_output.listing.items()])
    return BulkResult(pythonized_scalars, pythonized_list)


def bulkwalk(ip, community, oids, bulk_size=10, port=161,
             timeout=DEFAULT_TIMEOUT, version=Version.V2C):
    # type: (str, str, List[str], int, int, int) -> TWalkResponse
    """
    Delegates to :py:func:`~puresnmp.api.raw.bulkwalk` but returns simple
    Python types.

    See the "raw" equivalent for detailed documentation & examples.
    """

    result = multiwalk(
        ip, community, oids, port=port,
        fetcher=raw._bulkwalk_fetcher(  # pylint: disable=protected-access
            bulk_size),
        timeout=timeout, version=version)
    for oid, value in result:
        yield VarBind(oid, value)  # type: ignore


def table(ip, community, oid, port=161, num_base_nodes=0, version=Version.V2C):
    # type: (str, str, str, int, int) -> List[Dict[str, Any]]
    """
    Fetches a table from the SNMP agent. Each value will be converted to a
    pure-python type.

    See :py:func:`puresnmp.api.raw.table` for more information of the returned
    structure.
    """
    if num_base_nodes:
        warn(
            'Usage of "num_base_nodes" in table operations is no longer '
            "required",
            DeprecationWarning,
            stacklevel=2,
        )
    else:
        parsed_oid = OID(oid)
        num_base_nodes = len(parsed_oid) + 1  # type: ignore
    tmp = raw.table(ip, community, oid, port=port, num_base_nodes=num_base_nodes, version=version)
    output = []
    for row in tmp:
        index = row.pop('0')
        pythonized = {key: value.pythonize() for key, value in row.items()}
        pythonized['0'] = index
        output.append(pythonized)
    return output


def bulktable(ip, community, oid, port=161, num_base_nodes=0, bulk_size=10):
    # type: (str, str, str, int, int, int) -> List[Dict[str, Any]]
    """
    Fetch an SNMP table using "bulk" requests converting the values into pure
    Python types.

    See :py:func:`puresnmp.api.raw.table` for more information of the returned
    structure.

    .. versionadded: 1.7.0
    """
    if num_base_nodes:
        warn('Usage of "num_base_nodes" in table operations is no longer '
             'required', DeprecationWarning)
    else:
        parsed_oid = OID(oid)
        num_base_nodes = len(parsed_oid) + 1  # type: ignore
    tmp = raw.bulktable(ip, community, oid, port=port, bulk_size=bulk_size)
    output = []
    for row in tmp:
        index = row.pop('0')
        pythonized = {key: value.pythonize() for key, value in row.items()}
        pythonized['0'] = index
        output.append(pythonized)
    return output


def traps(listen_address='0.0.0.0', port=162, buffer_size=1024):
    # type: (str, int, int) -> Generator[TrapInfo, None, None]
    """
    A "pythonic" wrapper around :py:func:`puresnmp.api.raw.traps` output.
    """
    for raw_trap in raw.traps(listen_address, port, buffer_size):
        yield TrapInfo(raw_trap)



# pylint: skip-file

from ..x690 import types as t
from ..x690.util import TypeInfo
from .. import types as apptype

from . import ByteTester


UNIVERSAL = TypeInfo.UNIVERSAL
APPLICATION = TypeInfo.APPLICATION
PRIMITIVE = TypeInfo.PRIMITIVE
PRIVATE = TypeInfo.PRIVATE
CONSTRUCTED = TypeInfo.CONSTRUCTED
CONTEXT = TypeInfo.CONTEXT


def make_identifier_test(octet, expected_class, expected_pc, expected_value):
    def fun(self):
        result = TypeInfo.from_bytes(octet)
        expected = TypeInfo(expected_class, expected_pc, expected_value)
        self.assertEqual(result, expected)
    return fun


def add_class_detector(cls, expected_class, typeclass, tag):
    def fun(inst):
        result = t.Registry.get(typeclass, tag)
        inst.assertEqual(result, expected_class)
    fun.__name__ = 'test_%s' % expected_class.__name__
    setattr(cls, 'test_%s' % expected_class.__name__, fun)


class TestBasics(ByteTester):

    test_identifier_univ_prim = make_identifier_test(
        0b00000010, UNIVERSAL, PRIMITIVE, 0b00010)

    test_identifier_univ_const = make_identifier_test(
        0b00100010, UNIVERSAL, CONSTRUCTED, 0b00010)

    test_identifier_app_prim = make_identifier_test(
        0b01000010, APPLICATION, PRIMITIVE, 0b00010)

    test_identifier_app_const = make_identifier_test(
        0b01100010, APPLICATION, CONSTRUCTED, 0b00010)

    test_identifier_ctx_prim = make_identifier_test(
        0b10000010, CONTEXT, PRIMITIVE, 0b00010)

    test_identifier_ctx_const = make_identifier_test(
        0b10100010, CONTEXT, CONSTRUCTED, 0b00010)

    test_identifier_prv_prim = make_identifier_test(
        0b11000010, PRIVATE, PRIMITIVE, 0b00010)

    test_identifier_prv_const = make_identifier_test(
        0b11100010, PRIVATE, CONSTRUCTED, 0b00010)

    def test_identifier_long(self):
        self.skipTest('This is not yet implemented. I have not understood the '
                      'spec to confidently write a test')  # TODO


class TestClassDetector(ByteTester):
    pass


# "Standard" x690 types
add_class_detector(TestClassDetector, t.EOC, UNIVERSAL, 0x00)
add_class_detector(TestClassDetector, t.Boolean, UNIVERSAL, 0x01)
add_class_detector(TestClassDetector, t.Integer, UNIVERSAL, 0x02)
add_class_detector(TestClassDetector, t.BitString, UNIVERSAL, 0x03)
add_class_detector(TestClassDetector, t.OctetString, UNIVERSAL, 0x04)
add_class_detector(TestClassDetector, t.Null, UNIVERSAL, 0x05)
add_class_detector(TestClassDetector, t.ObjectIdentifier, UNIVERSAL, 0x06)
add_class_detector(TestClassDetector, t.ObjectDescriptor, UNIVERSAL, 0x07)
add_class_detector(TestClassDetector, t.External, UNIVERSAL, 0x08)
add_class_detector(TestClassDetector, t.Real, UNIVERSAL, 0x09)
add_class_detector(TestClassDetector, t.Enumerated, UNIVERSAL, 0x0A)
add_class_detector(TestClassDetector, t.EmbeddedPdv, UNIVERSAL, 0x0B)
add_class_detector(TestClassDetector, t.Utf8String, UNIVERSAL, 0x0C)
add_class_detector(TestClassDetector, t.RelativeOid, UNIVERSAL, 0x0D)
add_class_detector(TestClassDetector, t.Sequence, UNIVERSAL, 0x10)
add_class_detector(TestClassDetector, t.Set, UNIVERSAL, 0x11)
add_class_detector(TestClassDetector, t.NumericString, UNIVERSAL, 0x12)
add_class_detector(TestClassDetector, t.PrintableString, UNIVERSAL, 0x13)
add_class_detector(TestClassDetector, t.T61String, UNIVERSAL, 0x14)
add_class_detector(TestClassDetector, t.VideotexString, UNIVERSAL, 0x15)
add_class_detector(TestClassDetector, t.IA5String, UNIVERSAL, 0x16)
add_class_detector(TestClassDetector, t.UtcTime, UNIVERSAL, 0x17)
add_class_detector(TestClassDetector, t.GeneralizedTime, UNIVERSAL, 0x18)
add_class_detector(TestClassDetector, t.GraphicString, UNIVERSAL, 0x19)
add_class_detector(TestClassDetector, t.VisibleString, UNIVERSAL, 0x1a)
add_class_detector(TestClassDetector, t.GeneralString, UNIVERSAL, 0x1b)
add_class_detector(TestClassDetector, t.UniversalString, UNIVERSAL, 0x1c)
add_class_detector(TestClassDetector, t.CharacterString, UNIVERSAL, 0x1d)
add_class_detector(TestClassDetector, t.BmpString, UNIVERSAL, 0x1e)


# Application (SNMP-specific) Types
add_class_detector(TestClassDetector, apptype.IpAddress, APPLICATION, 0x00)
add_class_detector(TestClassDetector, apptype.Counter, APPLICATION, 0x01)
add_class_detector(TestClassDetector, apptype.Gauge, APPLICATION, 0x02)
add_class_detector(TestClassDetector, apptype.TimeTicks, APPLICATION, 0x03)
add_class_detector(TestClassDetector, apptype.Opaque, APPLICATION, 0x04)
add_class_detector(TestClassDetector, apptype.NsapAddress, APPLICATION, 0x05)
add_class_detector(TestClassDetector, apptype.Counter64, APPLICATION, 0x06)



'''
Tests for unit-test helpers
'''

from textwrap import dedent
from binascii import hexlify
import puresnmp.test as th
from io import StringIO


def test_readbytes_multiple():
    data = StringIO(dedent(
        u'''\
        #
        # This is a comment
        #

        30 2d 02 01 01 04 07 70  72 69 76 61 74 65 a2 1f   0-.....private..

        ----

        30 2d 02 01 01 04 07 70  72 69 76 61 74 65 a2 1f   0-.....private..

        ----

        30 2e 02 01 01 04 07 70  72 69 76 61 74 65 a2 20   0......private.
        '''
    ))
    expected = [
        b'\x30\x2d\x02\x01\x01\x04\x07\x70\x72\x69\x76\x61\x74\x65\xa2\x1f',
        b'\x30\x2d\x02\x01\x01\x04\x07\x70\x72\x69\x76\x61\x74\x65\xa2\x1f',
        b'\x30\x2e\x02\x01\x01\x04\x07\x70\x72\x69\x76\x61\x74\x65\xa2\x20',
    ]

    result = list(th.readbytes_multiple(data))

    assert result == expected


def test_readbytes_multiple_byte_positions():
    '''
    If ASCII columns are defined, we should still properly split the packets
    '''
    data = StringIO(dedent(
        u'''\
        # -*- ascii-cols: 6-56 -*-
        # Sequence of responses for an SNMP Walk operation which eventually falls into
        # a special "endOfMibView" marker (a varbind with binary *value* 0x82 0x00).

        0000: 30 37 02 01  01 04 07 70  72 69 76 61  74 65 A2 29    07.....private.)

        ----

        0000: 31 37 02 01  01 04 07 70  72 69 76 61  74 65 A2 29    07.....private.)
        '''
    ))

    expected = [
        b'\x30\x37\x02\x01\x01\x04\x07\x70\x72\x69\x76\x61\x74\x65\xA2\x29',
        b'\x31\x37\x02\x01\x01\x04\x07\x70\x72\x69\x76\x61\x74\x65\xA2\x29',
    ]

    result = list(th.readbytes_multiple(data))

    assert result == expected


def test_readbytes():
    data = StringIO(dedent(
        u'''\
        0000: 30 29 02 01  01 04 07 70  72 69 76 61  74 65 A5 1B    0).....private..
        '''
    ))

    expected = b'\x30\x29\x02\x01\x01\x04\x07\x70\x72\x69\x76\x61\x74\x65\xA5\x1B'
    result = th.readbytes(data)
    assert result == expected


def test_chunker():
    data = [1, 2, -1, 3, 4]
    expected = [[1, 2], [3, 4]]
    result = list(th.chunker(data, is_boundary=lambda x: x == -1))
    assert result == expected

    data = [-1, 1, 2, -1, 3, 4]
    expected = [[1, 2], [3, 4]]
    result = list(th.chunker(data, is_boundary=lambda x: x == -1))
    assert result == expected

    data = [1, 2, -1, 3, 4, -1]
    expected = [[1, 2], [3, 4]]
    result = list(th.chunker(data, is_boundary=lambda x: x == -1))
    assert result == expected

    data = [-1]
    expected = []
    result = list(th.chunker(data, is_boundary=lambda x: x == -1))
    assert result == expected

    data = []
    expected = []
    result = list(th.chunker(data, is_boundary=lambda x: x == -1))
    assert result == expected



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['puresnmp',
 'puresnmp.aio',
 'puresnmp.aio.api',
 'puresnmp.api',
 'puresnmp.test',
 'puresnmp.test.external_api',
 'puresnmp.test.gh-issues',
 'puresnmp.test.test',
 'puresnmp.test.x690',
 'puresnmp.x690']

package_data = \
{'': ['*'],
 'puresnmp.test': ['data/*', 'data/apiv1/*', 'data/gh-issues/*', 'data/x690/*']}

install_requires = \
['six', 't61codec>=1.0.1', 'verlib']

extras_require = \
{':python_version < "3.3"': ['ipaddress', 'mock'],
 ':python_version < "3.5"': ['typing'],
 ':python_version < "3.8"': ['importlib_metadata']}

setup_kwargs = {
    'name': 'puresnmp',
    'version': '1.11.0',
    'description': 'Pure Python SNMP implementation',
    'long_description': ".. >>> Shields >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n\n.. image:: https://travis-ci.org/exhuma/puresnmp.svg?branch=develop\n    :target: https://travis-ci.org/exhuma/puresnmp\n\n.. image:: https://readthedocs.org/projects/puresnmp/badge/?version=latest\n    :target: http://puresnmp.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\n.. image:: https://img.shields.io/badge/repository-github-green.svg?style=flat\n    :target: https://github.com/exhuma/puresnmp\n    :alt: Github Project\n\n.. image:: https://img.shields.io/pypi/v/puresnmp.svg\n    :alt: PyPI\n    :target: https://pypi.org/project/puresnmp/\n\n.. <<< Shields <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<\n\n----\n\n\nTL;DR\n-----\n\nJump right in with the `cookbook`_\n\n----\n\n\nQuick Info\n----------\n\nWhat\n    A pure Python implementation for Python 3.3+ of SNMP without any external\n    dependencies (neither MIBs or libsnmp).\n\nWhy\n    SNMP in itself is simple and well defined. A bit convoluted, but simple.\n    MIB parsing however complicates the code-base and is *technically* not\n    required. They add typing information and variables and give names to OIDs.\n    All existing libraries have a direct or indirect dependency on libsnmp.\n    With all the advantages and disadvantages.\n\n    The aim of this project is to focus on SNMP in itself and provide a very\n    simple API. Instead of implementing ASN.1 parsing, the SNMP related ASN.1\n    and X.690 information is hard-coded (keeping in mind that all that's\n    hard-coded is well defined).\n\n    It is of course possible to *wrap* this package in another package adding\n    MIB parsing and processing. This is, and will be however **out of the scope\n    of this project**!\n\nWhen\n    First commit: Sat Jul 23 12:01:05 2016 +0200\n\nWho\n    Michel Albert\n\n\nInstallation\n------------\n\n::\n\n    pip install puresnmp\n\n\n\nPackage Version Numbers\n-----------------------\n\nAs an important side-note, you might want to know that this project follows\n`Semantic Versioning`_.\n\nExamples\n--------\n\nSee the `cookbook`_.\n\n.. _cookbook: http://puresnmp.readthedocs.io/en/latest/cookbook/index.html\n.. _Semantic Versioning: http://semver.org/spec/v2.0.0.html\n",
    'author': 'Michel Albert',
    'author_email': 'michel@albert.lu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/exhuma/puresnmp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



# pylint: skip-file


"""
Test the external "raw" interface.

"raw" means that types are not converted to pythonic types but are returned as
PureSNMP object instances.
"""

from __future__ import print_function

import sys
import unittest
from datetime import timedelta
from logging import WARNING, Handler, getLevelName, getLogger
from unittest import skipUnless

import six

from puresnmp.api.raw import (
    bulkget,
    bulktable,
    bulkwalk,
    get,
    getnext,
    multiget,
    multiset,
    multiwalk,
    set,
    table,
    traps,
    walk,
)
from puresnmp.const import Version
from puresnmp.exc import FaultySNMPImplementation, NoSuchOID, SnmpError
from puresnmp.pdu import (
    BulkGetRequest,
    GetNextRequest,
    GetRequest,
    GetResponse,
    VarBind
)
from puresnmp.transport import SocketInfo, SocketResponse
from puresnmp.types import Counter, Gauge, IpAddress, TimeTicks
from puresnmp.util import BulkResult
from puresnmp.x690.types import (
    Integer,
    ObjectIdentifier,
    OctetString,
    Sequence,
    to_bytes
)

from . import ByteTester, CapturingHandler, readbytes, readbytes_multiple

try:
    from unittest.mock import patch, call
except ImportError:
    from mock import patch, call  # type: ignore



class TestGet(ByteTester):

    def test_get_string(self):
        data = readbytes('get_sysdescr_01.hex')
        expected = OctetString(
            b'Linux d24cf7f36138 4.4.0-28-generic #47-Ubuntu SMP '
            b'Fri Jun 24 10:09:13 UTC 2016 x86_64')
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            result = get('::1', 'private', '1.2.3')
        self.assertEqual(result, expected)

    def test_get_oid(self):
        data = readbytes('get_sysoid_01.hex')
        expected = ObjectIdentifier.from_string('1.3.6.1.4.1.8072.3.2.10')
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            result = get('::1', 'private', '1.2.3')
        self.assertEqual(result, expected)

    def test_get_multiple_return_binds(self):
        """
        A "GET" response should only return one varbind.
        """
        data = readbytes('get_sysoid_01_error.hex')
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            with six.assertRaisesRegex(self, SnmpError, 'varbind'):
                get('::1', 'private', '1.2.3')

    def test_get_non_existing_oid(self):
        """
        A "GET" response on a non-existing OID should raise an appropriate
        exception.
        """
        data = readbytes('get_non_existing.hex')
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            with self.assertRaises(NoSuchOID):
                get('::1', 'private', '1.2.3')


class TestWalk(unittest.TestCase):

    def test_walk(self):
        response_1 = readbytes('walk_response_1.hex')
        response_2 = readbytes('walk_response_2.hex')
        response_3 = readbytes('walk_response_3.hex')

        expected = [VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.5.1'),
            Gauge(10000000)
        ), VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.5.13'),
            Gauge(4294967295)
        )]

        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.side_effect = [response_1, response_2, response_3]
            mck().get_request_id.return_value = 0
            result = list(walk('::1', 'public', '1.3.6.1.2.1.2.2.1.5'))
        self.assertEqual(result, expected)

    def test_walk_multiple_return_binds(self):
        """
        A "WALK" response should only return one varbind.
        """
        data = readbytes('get_sysoid_01_error.hex')
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            with six.assertRaisesRegex(self, SnmpError, 'varbind'):
                next(walk('::1', 'private', '1.2.3'))


class TestMultiGet(unittest.TestCase):

    def test_multiget(self):
        data = readbytes('multiget_response.hex')
        expected = [
            ObjectIdentifier.from_string('1.3.6.1.4.1.8072.3.2.10'),
            OctetString(b"Linux 7fbf2f0c363d 4.4.0-28-generic "
                        b"#47-Ubuntu SMP Fri Jun 24 10:09:13 "
                        b"UTC 2016 x86_64")
        ]
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            result = multiget('::1', 'private', [
                '1.3.6.1.2.1.1.2.0',
                '1.3.6.1.2.1.1.1.0',
            ])
            result = list(result)
        self.assertEqual(result, expected)


class TestMultiWalk(unittest.TestCase):

    def test_multi_walk(self):
        response_1 = readbytes('multiwalk_response_1.hex')
        response_2 = readbytes('multiwalk_response_2.hex')
        response_3 = readbytes('multiwalk_response_3.hex')

        expected = [VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.1.1'), Integer(1)
        ), VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.2.1'),
            OctetString(b'lo')
        ), VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.1.78'), Integer(78)
        ), VarBind(
            ObjectIdentifier.from_string('1.3.6.1.2.1.2.2.1.2.78'),
            OctetString(b'eth0')
        )]

        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.side_effect = [response_1, response_2, response_3]
            mck().get_request_id.return_value = 0
            result = list(multiwalk('::1', 'public', [
                '1.3.6.1.2.1.2.2.1.1',
                '1.3.6.1.2.1.2.2.1.2'
            ]))
        self.assertEqual(result, expected)

    def test_multiwalk_non_containment(self):
        '''
        Running a multiwalk should raise an exception if the agent returns OIDs
        which are not properly increasing.
        '''
        from puresnmp.pdu import GetResponse
        OID = ObjectIdentifier.from_string

        # First case: Returned OIDs are the same
        response = Sequence(
            Integer(1),
            OctetString(b'public'),
            GetResponse(
                123,
                [
                    VarBind(oid=OID('1.2.3'), value=Integer(30)),
                    VarBind(oid=OID('2.3.4'), value=Integer(40)),
                ]
            )
        )
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.side_effect = [to_bytes(response)]
            mck().get_request_id.return_value = 0
            with self.assertRaises(FaultySNMPImplementation):
                list(multiwalk('::1', 'public', [
                    '1.2.3',
                    '2.3.4',
                ]))


        # Second case: Returned OIDs are smaller
        response = Sequence(
            Integer(1),
            OctetString(b'public'),
            GetResponse(
                123,
                [
                    VarBind(oid=OID('1.2.2'), value=Integer(30)),
                    VarBind(oid=OID('2.3.3'), value=Integer(40)),
                ]
            )
        )
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.side_effect = [to_bytes(response)]
            mck().get_request_id.return_value = 0
            with self.assertRaises(FaultySNMPImplementation):
                list(multiwalk('::1', 'public', [
                    '1.2.3',
                    '2.3.4',
                ]))

    def test_eom(self):
        '''
        A test for a walk operation which runs into the endOfMibView marker
        '''

        data_generator = readbytes_multiple('x690/multiwalk_endofmibview.hex')

        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.side_effect = data_generator
            mck().get_request_id.return_value = 0
            result = multiwalk(u'::1', u'public', [
                u'1.3.6.1.6.3.16.1.5.2.1.6.6.95.110.111.110.101.95.1',
            ])
            result = list(result)

        OID = ObjectIdentifier.from_string
        root = '1.3.6.1.6.3.16.1.5.2.1.6.6.95.110.111.110.101.95.1.'
        expected = [
            VarBind(OID(root+'0'), Integer(1)),
            VarBind(OID(root+'1'), Integer(1)),
            VarBind(OID(root+'2'), Integer(1)),
        ]

        self.assertEqual(result, expected)


class TestMultiSet(unittest.TestCase):

    def test_multiset(self):
        """
        Test setting multiple OIDs at once.

        NOTE: The OID '1.3.6.1.2.1.1.5.0' below is manually edited for
              unit-testing. It probably has a different type in the real world!
        """
        data = readbytes('multiset_response.hex')
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            result = multiset('::1', 'private', [
                ('1.3.6.1.2.1.1.4.0', OctetString(b'hello@world.com')),
                ('1.3.6.1.2.1.1.5.0', OctetString(b'hello@world.com')),
            ])
        expected = {
            '1.3.6.1.2.1.1.4.0': OctetString(b'hello@world.com'),
            '1.3.6.1.2.1.1.5.0': OctetString(b'hello@world.com'),
        }
        self.assertEqual(result, expected)


class TestGetNext(unittest.TestCase):

    def test_get_call_args(self):
        data = readbytes('dummy.hex')  # any dump would do
        packet = Sequence(
            Integer(Version.V2C),
            OctetString('public'),
            GetNextRequest(0, ObjectIdentifier(1, 2, 3))
        )
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            getnext('::1', 'public', '1.2.3')
            mck.assert_called_with(timeout=6)
            mck().send.assert_called_with('::1', 161, to_bytes(packet))

    def test_getnext(self):
        data = readbytes('getnext_response.hex')
        expected = VarBind('1.3.6.1.6.3.1.1.6.1.0', Integer(354522558))

        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            result = getnext('::1', 'private', '1.3.6.1.5')
        self.assertEqual(result, expected)

    def test_getnext_increasing_oid_strict(self):
        '''
        When running "getnext" we expect a different OID than the one we passed
        in. If not, this can cause endless-loops in the worst case. Faulty SNMP
        implementations may behave this way!
        '''
        requested_oid = ObjectIdentifier(1, 2, 3, 4)
        response_object = Sequence(
            Integer(1),
            OctetString(b'public'),
            GetResponse(
                234,
                [VarBind(requested_oid, Integer(123))]
            )
        )
        response_bytes = to_bytes(response_object)

        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = response_bytes
            mck().get_request_id.return_value = 0
            with self.assertRaises(FaultySNMPImplementation):
                getnext('::1', 'private', '1.2.3.4')

    def test_walk_increasing_oid_lenient(self):
        '''
        We want to be able to allow faulty SNMP implementations to at least try
        to fetch the values in a walk which are not increasing. It should read
        up to the values which are no longer increasing and emit a warning.
        '''
        logger = getLogger('puresnmp')
        handler = CapturingHandler()
        logger.addHandler(handler)

        response_binds = [
            VarBind(ObjectIdentifier(1, 2, 3), Integer(123)),
            VarBind(ObjectIdentifier(1, 2, 4), Integer(124)),
            VarBind(ObjectIdentifier(1, 2, 5), Integer(125)),
            VarBind(ObjectIdentifier(1, 2, 1), Integer(121)),  # non-increasing
        ]
        response_packets = [
            Sequence(
                Integer(1),
                OctetString(b'public'),
                GetResponse(
                    234,
                    [bind]
                )
            )
            for bind in response_binds
        ]
        response_bytes = [to_bytes(packet) for packet in response_packets]

        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.side_effect = response_bytes
            mck().get_request_id.return_value = 0
            result = list(walk('::1', 'private', '1.2', errors='warn'))


        # The last OID in the mocked responses is decreasing so we want to read
        # just up to that point.
        expected = [
            VarBind(ObjectIdentifier(1, 2, 3), Integer(123)),
            VarBind(ObjectIdentifier(1, 2, 4), Integer(124)),
            VarBind(ObjectIdentifier(1, 2, 5), Integer(125)),
        ]
        self.assertEqual(result, expected)

        # We also want to make sure that we have a proper warning about this
        handler.assertContains(WARNING, r'.*1.2.1.*1.2.5.*')
        logger.removeHandler(handler)


    def test_walk_endless_loop(self):
        '''
        In rare cases, some devices fall into an endless loop by returning the
        requested OID on a "getnext" call during a "walk" operation. A SNMP
        client behaving according to the SNMP spec will fall into an endless
        loop. This test fakes such a case and revents the loop.
        '''
        response_binds = [
            VarBind(ObjectIdentifier(1, 2, 3), Integer(123)),
            VarBind(ObjectIdentifier(1, 2, 4), Integer(124)),
            VarBind(ObjectIdentifier(1, 2, 5), Integer(125)),
            VarBind(ObjectIdentifier(1, 2, 5), Integer(125)),  # same OID
            VarBind(ObjectIdentifier(1, 2, 5), Integer(125)),  # same OID
            VarBind(ObjectIdentifier(1, 2, 5), Integer(125)),  # same OID
        ]
        response_packets = [
            Sequence(
                Integer(1),
                OctetString(b'public'),
                GetResponse(
                    234,
                    [bind]
                )
            )
            for bind in response_binds
        ]
        response_bytes = [to_bytes(packet) for packet in response_packets]

        handler = CapturingHandler()
        logger = getLogger('puresnmp')
        logger.addHandler(handler)
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.side_effect = response_bytes
            mck().get_request_id.return_value = 0
            result = list(walk('::1', 'private', '1.2', errors='warn'))
        logger.removeHandler(handler)


        # The last OID in the mocked responses is decreasing so we want to read
        # just up to that point.
        expected = [
            VarBind(ObjectIdentifier(1, 2, 3), Integer(123)),
            VarBind(ObjectIdentifier(1, 2, 4), Integer(124)),
            VarBind(ObjectIdentifier(1, 2, 5), Integer(125)),
        ]
        self.assertEqual(result, expected)

        # We also want to make sure that we have a proper warning about this
        handler.assertContains(WARNING, r'.*1.2.5.*')


class TestGetBulkGet(unittest.TestCase):

    def test_get_call_args(self):
        data = readbytes('dummy.hex')  # any dump would do
        packet = Sequence(
            Integer(Version.V2C),
            OctetString('public'),
            BulkGetRequest(0, 1, 2,
                           ObjectIdentifier(1, 2, 3),
                           ObjectIdentifier(1, 2, 4))
        )
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            bulkget('::1', 'public',
                    ['1.2.3'],
                    ['1.2.4'],
                    max_list_size=2)
            mck.assert_called_with(timeout=6)
            mck().send.assert_called_with('::1', 161, to_bytes(packet))

    def test_bulkget(self):
        data = readbytes('bulk_get_response.hex')
        expected = BulkResult(
            {'1.3.6.1.2.1.1.1.0': OctetString(
                b'Linux 7e68e60fe303 4.4.0-28-generic '
                b'#47-Ubuntu SMP Fri Jun 24 10:09:13 UTC 2016 x86_64')},
            {'1.3.6.1.2.1.3.1.1.1.10.1.172.17.0.1': Integer(10),
             '1.3.6.1.2.1.3.1.1.2.10.1.172.17.0.1':
             OctetString(b'\x02B\xe2\xc5\x8d\t'),
             '1.3.6.1.2.1.3.1.1.3.10.1.172.17.0.1':
             IpAddress(b'\xac\x11\x00\x01'),
             '1.3.6.1.2.1.4.1.0': Integer(1),
             '1.3.6.1.2.1.4.3.0': Counter(57)})

        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            result = bulkget('::1', 'public',
                             ['1.3.6.1.2.1.1.1'],
                             ['1.3.6.1.2.1.3.1'],
                             max_list_size=5)
        self.assertEqual(result, expected)

    def test_eom(self):
        '''
        Test a bulg-get operation which runs into the "endOfMibView" marker.
        '''

        data = readbytes('x690/bulk_get_eom_response.hex')
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0
            result = bulkget('::1', 'public', [], ['1.2.4'], max_list_size=10)

        expected_scalars = {}
        self.assertEqual(result.scalars, expected_scalars)

        root = '1.3.6.1.6.3.16.1.5.2.1.6.6.95.110.111.110.101.95.1.'
        expected_listing = {
            root + '0': Integer(1),
            root + '1': Integer(1),
            root + '2': Integer(1),
        }

        self.assertEqual(result.listing, expected_listing)


class TestGetBulkWalk(unittest.TestCase):

    def test_get_call_args(self):
        data = readbytes('dummy.hex')  # any dump would do
        packet = Sequence(
            Integer(Version.V2C),
            OctetString('public'),
            BulkGetRequest(0, 0, 2, ObjectIdentifier(1, 2, 3))
        )
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0

            # we need to wrap this in a list to consume the generator.
            list(bulkwalk('::1', 'public',
                          ['1.2.3'],
                          bulk_size=2))
            mck.assert_called_with(timeout=6)
            mck().send.assert_called_with('::1', 161, to_bytes(packet))

    def test_get_call_args_issue_22(self):
        data = readbytes('dummy.hex')  # any dump would do
        packet = Sequence(
            Integer(Version.V2C),
            OctetString('public'),
            BulkGetRequest(0, 0, 2, ObjectIdentifier(1, 2, 3))
        )
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().send.return_value = data
            mck().get_request_id.return_value = 0

            with six.assertRaisesRegex(self, TypeError, 'OIDS.*list'):
                # we need to wrap this in a list to consume the generator.
                list(bulkwalk('::1', 'public', '1.2.3', bulk_size=2))

    @patch('puresnmp.api.raw.Transport')
    def test_bulkwalk(self, mck_transport):
        req1 = readbytes('bulkwalk_request_1.hex')
        req2 = readbytes('bulkwalk_request_2.hex')
        req3 = readbytes('bulkwalk_request_3.hex')

        responses = [
            readbytes('bulkwalk_response_1.hex'),
            readbytes('bulkwalk_response_2.hex'),
            readbytes('bulkwalk_response_3.hex'),
        ]
        mck_transport().send.side_effect = responses

        request_ids = [1001613222, 1001613223, 1001613224]
        mck_transport().get_request_id.side_effect = request_ids

        result = list(bulkwalk('127.0.0.1', 'private', ['1.3.6.1.2.1.2.2'],
                               bulk_size=20))

        mck_transport.assert_called_with(timeout=6)
        self.assertEqual(mck_transport().send.mock_calls, [
            call('127.0.0.1', 161, req1),
            call('127.0.0.1', 161, req2),
            call('127.0.0.1', 161, req3),
        ])

        expected = [
            VarBind('1.3.6.1.2.1.2.2.1.1.1', Integer(1)),
            VarBind('1.3.6.1.2.1.2.2.1.1.10', Integer(10)),
            VarBind('1.3.6.1.2.1.2.2.1.2.1', OctetString(b"lo")),
            VarBind('1.3.6.1.2.1.2.2.1.2.10', OctetString(b"eth0")),
            VarBind('1.3.6.1.2.1.2.2.1.3.1', Integer(24)),
            VarBind('1.3.6.1.2.1.2.2.1.3.10', Integer(6)),
            VarBind('1.3.6.1.2.1.2.2.1.4.1', Integer(65536)),
            VarBind('1.3.6.1.2.1.2.2.1.4.10', Integer(1500)),
            VarBind('1.3.6.1.2.1.2.2.1.5.1',  Gauge(10000000)),
            VarBind('1.3.6.1.2.1.2.2.1.5.10',  Gauge(4294967295)),
            VarBind('1.3.6.1.2.1.2.2.1.6.1', OctetString(b"")),
            VarBind('1.3.6.1.2.1.2.2.1.6.10',
                OctetString(b"\x02\x42\xAC\x11\x00\x02")),
            VarBind('1.3.6.1.2.1.2.2.1.7.1', Integer(1)),
            VarBind('1.3.6.1.2.1.2.2.1.7.10', Integer(1)),
            VarBind('1.3.6.1.2.1.2.2.1.8.1', Integer(1)),
            VarBind('1.3.6.1.2.1.2.2.1.8.10', Integer(1)),
            VarBind('1.3.6.1.2.1.2.2.1.9.1', TimeTicks(0)),
            VarBind('1.3.6.1.2.1.2.2.1.9.10', TimeTicks(0)),
            VarBind('1.3.6.1.2.1.2.2.1.10.1', Counter(172)),
            VarBind('1.3.6.1.2.1.2.2.1.10.10', Counter(60558)),
            VarBind('1.3.6.1.2.1.2.2.1.11.1', Counter(2)),
            VarBind('1.3.6.1.2.1.2.2.1.11.10', Counter(564)),
            VarBind('1.3.6.1.2.1.2.2.1.12.1', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.12.10', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.13.1', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.13.10', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.14.1', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.14.10', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.15.1', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.15.10', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.16.1', Counter(172)),
            VarBind('1.3.6.1.2.1.2.2.1.16.10', Counter(44295)),
            VarBind('1.3.6.1.2.1.2.2.1.17.1', Counter(2)),
            VarBind('1.3.6.1.2.1.2.2.1.17.10', Counter(442)),
            VarBind('1.3.6.1.2.1.2.2.1.18.1', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.18.10', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.19.1', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.19.10', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.20.1', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.20.10', Counter(0)),
            VarBind('1.3.6.1.2.1.2.2.1.21.1',  Gauge(0)),
            VarBind('1.3.6.1.2.1.2.2.1.21.10',  Gauge(0)),
            VarBind('1.3.6.1.2.1.2.2.1.22.1', ObjectIdentifier(0, 0)),
            VarBind('1.3.6.1.2.1.2.2.1.22.10', ObjectIdentifier(0, 0))
        ]
        self.assertEqual(result, expected)


class TestTable(unittest.TestCase):

    @patch('puresnmp.api.raw.walk')
    def test_table(self, mck_walk):
        mck_walk.return_value = [
            VarBind('1.2.1.1.1', OctetString(b'row 1 col 1')),
            VarBind('1.2.1.1.2', OctetString(b'row 2 col 1')),
            VarBind('1.2.1.2.1', OctetString(b'row 1 col 2')),
            VarBind('1.2.1.2.2', OctetString(b'row 2 col 2')),
        ]
        result = table('192.0.2.1', 'private', '1.2')
        expected = [
            {'0': '1',
             '1': OctetString('row 1 col 1'),
             '2': OctetString('row 1 col 2')},
            {'0': '2',
             '1': OctetString('row 2 col 1'),
             '2': OctetString('row 2 col 2')},
        ]
        self.assertEqual(sorted(result, key=lambda x: x['0']), expected)

    @patch('puresnmp.api.raw.walk')
    def test_table_complex_row_id(self, mck_walk):
        mck_walk.return_value = [
            VarBind('1.2.1.1.1.1', OctetString(b'row 1.1.1 col 1')),
            VarBind('1.2.1.2.1.1', OctetString(b'row 2.1.1 col 1')),
            VarBind('1.2.2.1.1.1', OctetString(b'row 1.1.1 col 2')),
            VarBind('1.2.2.2.1.1', OctetString(b'row 2.1.1 col 2')),
        ]
        result = table('192.0.2.1', 'private', '1.2', num_base_nodes=2)
        expected = [
            {'0': '1.1.1',
             '1': OctetString('row 1.1.1 col 1'),
             '2': OctetString('row 1.1.1 col 2')},
            {'0': '2.1.1',
             '1': OctetString('row 2.1.1 col 1'),
             '2': OctetString('row 2.1.1 col 2')},
        ]
        self.assertEqual(sorted(result, key=lambda x: x['0']), expected)

    @patch('puresnmp.api.raw.walk')
    def test_table_base_oid(self, mck_walk):
        """
        The "table" function should be capable of detecting the
        "num_base_nodes" value by itself
        """
        mck_walk.return_value = [
            VarBind('1.2.1.1.1.1.1', OctetString(b'row 1.1.1 col 1')),
            VarBind('1.2.1.1.2.1.1', OctetString(b'row 2.1.1 col 1')),
            VarBind('1.2.1.2.1.1.1', OctetString(b'row 1.1.1 col 2')),
            VarBind('1.2.1.2.2.1.1', OctetString(b'row 2.1.1 col 2')),
        ]
        result = table('192.0.2.1', 'private', '1.2')
        expected = [
            {'0': '1.1.1',
             '1': OctetString('row 1.1.1 col 1'),
             '2': OctetString('row 1.1.1 col 2')},
            {'0': '2.1.1',
             '1': OctetString('row 2.1.1 col 1'),
             '2': OctetString('row 2.1.1 col 2')},
        ]
        self.assertEqual(sorted(result, key=lambda x: x['0']), expected)


class TestBulkTable(unittest.TestCase):

    @patch('puresnmp.api.raw.Transport')
    def test_bulktable(self, mck_transport):
        responses = readbytes_multiple('bulktable_response.hex')
        mck_transport().send.side_effect = responses

        request_ids = [1, 2, 3, 4, 5]
        mck_transport().get_request_id.side_effect = request_ids

        result = list(bulktable('127.0.0.1', 'private', '1.3.6.1.2.1.2.2'))

        expected = [{
            '0': '1',
            '1': Integer(1),
            '2': OctetString(b'lo'),
            '3': Integer(24),
            '4': Integer(65536),
            '5': Gauge(10000000),
            '6': OctetString(b''),
            '7': Integer(1),
            '8': Integer(1),
            '9': TimeTicks(0),
            '10': Counter(172),
            '11': Counter(2),
            '12': Counter(0),
            '13': Counter(0),
            '14': Counter(0),
            '15': Counter(0),
            '16': Counter(172),
            '17': Counter(2),
            '18': Counter(0),
            '19': Counter(0),
            '20': Counter(0),
            '21': Gauge(0),
            '22': ObjectIdentifier(0, 0),
        }, {
            '0': '4',
            '1': Integer(4),
            '2': OctetString(b'eth0'),
            '3': Integer(6),
            '4': Integer(1500),
            '5': Gauge(4294967295),
            '6': OctetString(b'\x02B\xac\x11\x00\x02'),
            '7': Integer(1),
            '8': Integer(1),
            '9': TimeTicks(0),
            '10': Counter(548760),
            '11': Counter(3888),
            '12': Counter(0),
            '13': Counter(0),
            '14': Counter(0),
            '15': Counter(0),
            '16': Counter(186660),
            '17': Counter(1875),
            '18': Counter(0),
            '19': Counter(0),
            '20': Counter(0),
            '21': Gauge(0),
            '22': ObjectIdentifier(0, 0),
        }]
        self.assertEqual(sorted(result, key=lambda x: x['0']), expected)


class TestTraps(unittest.TestCase):

    def test_traps(self):
        data_generator = readbytes_multiple("trap_requests.hex")

        def socket_response_generator():
            for blob in data_generator:
                yield SocketResponse(blob, SocketInfo("192.0.2.1", 64001))

        expected = [
            VarBind(ObjectIdentifier((1, 3, 6, 1, 2, 1, 1, 3, 0)),
                    TimeTicks(794602)),
            VarBind(ObjectIdentifier((1, 3, 6, 1, 6, 3, 1, 1, 4, 1, 0)),
                    ObjectIdentifier((1, 3, 6, 1, 4, 1, 8072, 2, 3, 0, 1))),
            VarBind(ObjectIdentifier((1, 3, 6, 1, 4, 1, 8072, 2, 3, 2, 1)),
                    Integer(123456)),
            VarBind(ObjectIdentifier((1, 3, 6, 1, 2, 1, 1, 3, 0)),
                    TimeTicks(795345)),
            VarBind(ObjectIdentifier((1, 3, 6, 1, 6, 3, 1, 1, 4, 1, 0)),
                    ObjectIdentifier((1, 3, 6, 1, 4, 1, 8072, 2, 3, 0, 1))),
            VarBind(ObjectIdentifier((1, 3, 6, 1, 4, 1, 8072, 2, 3, 2, 1)),
                    Integer(123457)),
            VarBind(ObjectIdentifier((1, 3, 6, 1, 2, 1, 1, 3, 0)),
                    TimeTicks(795538)),
            VarBind(ObjectIdentifier((1, 3, 6, 1, 6, 3, 1, 1, 4, 1, 0)),
                    ObjectIdentifier((1, 3, 6, 1, 4, 1, 8072, 2, 3, 0, 1))),
            VarBind(ObjectIdentifier((1, 3, 6, 1, 4, 1, 8072, 2, 3, 2, 1)),
                    Integer(123459)),
        ]
        result = []
        with patch('puresnmp.api.raw.Transport') as mck:
            mck().listen.return_value = socket_response_generator()
            for trap in traps():
                result.extend(trap.varbinds)
        self.assertEqual(result, expected)

    def test_traps_origin(self):
        """
        We want to see where a trap was sent from
        """
        data_generator = readbytes_multiple("trap_requests.hex")
        def socket_response_generator():
            for blob in data_generator:
                yield SocketResponse(blob, SocketInfo("192.0.2.1", 64001))
        # As we import "set" above we are no longer able to create empty sets -_-
        result = {1}
        with patch("puresnmp.api.raw.Transport") as mck:
            mck().listen.return_value = socket_response_generator()
            for trap in traps():
                result.add(trap.source)
        expected = {1, SocketInfo("192.0.2.1", 64001)}
        self.assertEqual(result, expected)



# type: ignore
# Type-Hinting is done in a stub file
"""
Overview
========

This module contains the encoding/decoding logic for data types as defined in
:term:`X.690`.

Each type is made available via a :py:class:`~.Registry` and can be retrieved
via :py:meth:`~.Registry.get`.

Additionally, given a :py:class:`bytes` object, the :py:func:`~.pop_tlv`
function can be used to parse theat bytes object and return a typed instance
from it. See :py:func:`~.pop_tlv` for details about it's behaviour!

.. note::
    The individual type classes in this module do not contain any additional
    documentation. The bulk of this module is documented in :py:class:`~.Type`
    and :py:class:`~.Registry`.

    For the rest, the type classes simply define the type identifier tag.

Supporting Additional Classes
=============================

Just by subclassing :py:class:`~.Type` and setting correct ``TAG`` and
``TYPECLASS`` values, most of the basic functionality will be covered by the
superclass. Type detection, and addition to the registry is automatic.
Subclassing is enough.

By default, a new type which does not override any methods will have it's value
reported as bytes objects. You may want to override at least
:py:meth:`~.Type.pythonize` to expose the value to users of the library as pure
Python objects.

Depending on type, you may also want to override certain methods. See
:py:class:`~.Sequence` and :py:class:`~.Integer` for more complex examples.
"""
# pylint: disable=abstract-method, missing-docstring

from __future__ import division, print_function, unicode_literals

import logging
import warnings

import six
import t61codec
from six.moves import zip_longest

from .exc import InvalidValueLength
from .util import (
    TypeInfo,
    decode_length,
    encode_length,
    int_from_bytes,
    to_bytes
)

try:
    unicode
except NameError:
    # pylint: disable=invalid-name
    unicode = str


LOG = logging.getLogger(__name__)


class Registry(type):

    __registry = {}

    def __new__(cls, name, parents, dict_):
        new_cls = super(Registry, cls).__new__(cls, name, parents, dict_)
        if hasattr(new_cls, 'TAG') and hasattr(new_cls, 'TYPECLASS'):
            Registry.__registry[(new_cls.TYPECLASS, new_cls.TAG)] = new_cls
        else:
            LOG.warning('Ignoring class %r. It is missing either the '
                        'TYPECLASS or TAG class attribute!', new_cls)
        return new_cls

    @staticmethod
    def get(typeclass, typeid):
        return Registry.__registry[(typeclass, typeid)]


def pop_tlv(data):
    """
    Given a :py:class:`bytes` object, inspects and parses the first octets (as
    many as required) to determine variable type (and corresponding Python
    class), and length. The class is then used to parse the *first* object in
    ``data``.  *data* itself will not be modified. Instead, a new modified copy
    of *data* is returned alongside the parsed object. This new object is the
    remainder after popping off the first object.

    Example::

        >>> data = b'\\x02\\x01\\x05\\x11'
        >>> pop_tlv(data)
        (Integer(5), b'\\x11')

    Note that in the example above, ``\\x11`` is the remainder of the bytes
    object after popping of the integer object.
    """
    # TODO: This function should be moved to another module (util maybe?).
    if not data:
        return Null(), b''
    type_ = TypeInfo.from_bytes(data[0])
    length, remainder = decode_length(data[1:])

    # determine how many octets are used to encode the length!
    offset = len(data) - len(remainder)
    chunk = data[:length+offset]
    try:
        cls = Registry.get(type_.cls, type_.tag)
        value = cls.from_bytes(chunk)
    except KeyError:
        # Add context information
        value = UnknownType.from_bytes(chunk)
    return value, remainder[length:]


@six.add_metaclass(Registry)
class Type(object):
    """
    The superclass for all supported types.
    """
    TYPECLASS = TypeInfo.UNIVERSAL
    TAG = 0
    value = None

    @classmethod
    def validate(cls, data):
        """
        Given a bytes object, checks if the given class *cls* supports decoding
        this object. If not, raises a ValueError.
        """
        # TODO: Making this function return a boolean instead of raising an exception would make the code potentially more readable.
        tinfo = TypeInfo.from_bytes(data[0])
        if tinfo.cls != cls.TYPECLASS or tinfo.tag != cls.TAG:
            raise ValueError('Invalid type header! '
                             'Expected a %s class with tag '
                             'ID 0x%02x, but got a %s class with '
                             'tag ID 0x%02x' % (
                                 cls.TYPECLASS, cls.TAG, tinfo.cls,
                                 six.byte2int(data)))

    @classmethod
    def from_bytes(cls, data):
        """
        Given a bytes object, this method reads the type information and length
        and uses it to convert the bytes representation into a python object.

        :raises puresnmp.x690.exc.InvalidValueLength: If the packet-length does
            not match the expected length reported in the package header.
        """

        if not data:
            return Null()
        cls.validate(data)
        expected_length, data = decode_length(data[1:])
        if len(data) != expected_length:
            raise InvalidValueLength(
                'Corrupt packet: Unexpected length for {0} Expected {1} '
                '(0x{1:02x}) but got {2} (0x{2:02x}). Consider increasing '
                'puresnmp.transport.BUFFER_SIZE.'.format(
                    cls, expected_length, len(data)))
        return cls.decode(data)

    @classmethod
    def decode(cls, data):  # pragma: no cover
        """
        This method takes a bytes object which contains the raw content octets
        of the object. That means, the octets *without* the type information
        and length.

        This function must be overridden by the concrete subclasses.
        """
        raise NotImplementedError(
            'Decoding is not yet implemented on %s' % cls)

    def __bytes__(self):  # pragma: no cover
        """
        Convert this instance into a bytes object. This must be implemented by
        subclasses.
        """
        raise NotImplementedError('Not yet implemented')

    def __repr__(self):
        # pylint: disable=no-member
        return '%s(%r)' % (self.__class__.__name__, self.value)

    def pythonize(self):
        """
        Convert this instance to an appropriate pure Python object.
        """
        # pylint: disable=no-member
        return self.value

    def pretty(self):  # pragma: no cover
        """
        Returns a readable representation (possibly multiline) of the value.

        By default this simply returns the string representation. But more
        complex values may override this.
        """
        return unicode(self)

    if six.PY2:  # pragma: no cover
        def __unicode__(self):
            return repr(self)

        def __str__(self):
            return self.__bytes__()


class UnknownType(Type):
    """
    A fallback type for anything not in X.690.

    Instances of this class contain the raw information as parsed from the
    bytes as the following attributes:

    * ``value``: The value without leading metadata (as bytes value)
    * ``tag``: The *unparsed* "tag". This is the type ID as defined in the
      reference document. See :py:class:`~puresnmp.x690.util.TypeInfo` for
      details.
    * ``typeinfo``: unused (derived from *tag* and only here for consistency
      with ``__repr__`` of this class).
    """

    value = b''

    def __init__(self, tag, value, typeinfo=None):
        self.value = value
        self.tag = tag
        self.length = len(value)
        self.typeinfo = TypeInfo.from_bytes(tag)

    def __bytes__(self):
        return to_bytes([self.tag]) + encode_length(self.length) + self.value

    def __repr__(self):
        return '%s(%r, %r, typeinfo=%r)' % (
            self.__class__.__name__, self.tag, self.value, self.typeinfo)

    def __eq__(self, other):
        # pylint: disable=unidiomatic-typecheck
        return (type(self) == type(other) and
                self.value == other.value and
                self.tag == other.tag)

    @classmethod
    def from_bytes(cls, data):
        """
        Overrides typical conversion by removing type validation. As, by
        definition this class is used for unknown types, we cannot validate
        them.
        """
        if not data:
            return Null()
        tag = six.byte2int(data)
        expected_length, data = decode_length(data[1:])
        if len(data) != expected_length:
            raise ValueError('Corrupt packet: Unexpected length for {0} '
                             'Expected {1} (0x{1:02x}) '
                             'but got {2} (0x{2:02x})'.format(
                                 UnknownType, expected_length, len(data)))
        return UnknownType(tag, data)


class NonASN1Type(UnknownType):  # pragma: no cover
    def __init__(self, tag, value):
        warnings.warn('puresnmp.x690.types.NonASN1Type is deprecated,'
                      ' replace it with UnknownType', stacklevel=2)
        super(NonASN1Type, self).__init__(tag, value)


class Boolean(Type):
    TAG = 0x01
    value = False

    @staticmethod
    def decode(data):
        return Boolean(data != b'\x00')

    @classmethod
    def validate(cls, data):
        super(Boolean, cls).validate(data)
        if six.indexbytes(data, 1) != 1:
            raise ValueError('Unexpected Boolean value. Length should be 1,'
                             ' it was %d' % six.indexbytes(data, 1))

    def __init__(self, value):
        self.value = value

    def __bytes__(self):
        return to_bytes([1, 1, int(self.value)])

    def __eq__(self, other):
        # pylint: disable=unidiomatic-typecheck
        return type(self) == type(other) and self.value == other.value


class Null(Type):
    TAG = 0x05

    def __init__(self):
        self.value = None

    @classmethod
    def validate(cls, data):
        super(Null, cls).validate(data)
        if six.indexbytes(data, 1) != 0:
            raise ValueError('Unexpected NULL value. Length should be 0, it '
                             'was %d' % six.indexbytes(data, 1))

    @classmethod
    def decode(cls, data):
        return Null()

    def __bytes__(self):
        return b'\x05\x00'

    def __eq__(self, other):
        # pylint: disable=unidiomatic-typecheck
        return type(self) == type(other)

    def __repr__(self):
        return 'Null()'

    def __bool__(self):
        return False

    def __nonzero__(self):  # pragma: no cover (__bool__ for py2)
        return False


class OctetString(Type):
    TAG = 0x04
    value = b''

    @classmethod
    def decode(cls, data):
        return cls(data)

    def __init__(self, value):
        if isinstance(value, unicode):
            self.value = value.encode('ascii')
        else:
            self.value = value
        self.length = encode_length(len(self.value))

    def __bytes__(self):
        tinfo = TypeInfo(self.TYPECLASS, TypeInfo.PRIMITIVE, self.TAG)
        return to_bytes(tinfo) + self.length + self.value

    def __eq__(self, other):
        # pylint: disable=unidiomatic-typecheck
        return type(self) == type(other) and self.value == other.value

    def pythonize(self):
        """
        Convert this object in an appropriate python object
        """
        return self.value


class Sequence(Type):
    """
    Represents an X.690 sequence type. Instances of this class are iterable and
    indexable.
    """
    TAG = 0x10
    value = []

    @classmethod
    def decode(cls, data):
        output = []
        while data:
            value, data = pop_tlv(data)
            output.append(value)
        return Sequence(*output)

    def __init__(self, *items):
        self.items = items

    def __bytes__(self):
        items = [to_bytes(item) for item in self]
        output = b''.join(items)
        length = encode_length(len(output))
        tinfo = TypeInfo(TypeInfo.UNIVERSAL, TypeInfo.CONSTRUCTED,
                         Sequence.TAG)
        return to_bytes(tinfo) + length + output

    def __eq__(self, other):
        # pylint: disable=unidiomatic-typecheck
        return type(self) == type(other) and self.items == other.items

    def __repr__(self):
        item_repr = [repr(item) for item in self]
        return 'Sequence(%s)' % ', '.join(item_repr)

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return iter(self.items)

    def __getitem__(self, idx):
        return self.items[idx]

    def pythonize(self):
        return [obj.pythonize() for obj in self]

    def pretty(self):  # pragma: no cover
        """
        Overrides :py:meth:`.Type.pretty`
        """
        lines = [self.__class__.__name__]
        for item in self.items:
            lines.append('   %s' % item.pretty())
        return '\n'.join(lines)


class Integer(Type):
    SIGNED = True
    TAG = 0x02
    value = 0

    @classmethod
    def decode(cls, data):
        return cls(int_from_bytes(data, 'big', signed=cls.SIGNED))

    def __init__(self, value):
        self.value = value

    def __bytes__(self):
        octets = [self.value & 0b11111111]

        # Append remaining octets for long integers.
        remainder = self.value
        while remainder not in (0, -1):
            remainder = remainder >> 8
            octets.append(remainder & 0b11111111)

        if remainder == 0 and octets[-1] == 0b10000000:
            octets.append(0)
        octets.reverse()

        # remove leading octet if there is a string of 9 zeros or ones
        while (len(octets) > 1 and
               ((octets[0] == 0 and octets[1] & 0b10000000 == 0) or
                (octets[0] == 0b11111111 and octets[1] & 0b10000000 != 0))):
            del octets[0]

        tinfo = TypeInfo(self.TYPECLASS, TypeInfo.PRIMITIVE, self.TAG)
        return to_bytes(tinfo) + to_bytes([len(octets)]) + to_bytes(octets)

    def __eq__(self, other):
        # pylint: disable=unidiomatic-typecheck
        return type(self) == type(other) and self.value == other.value


class ObjectIdentifier(Type):
    """
    Represents an OID.

    Instances of this class support containment checks to determine if one OID
    is a sub-item of another::

        >>> ObjectIdentifier(1, 2, 3, 4, 5) in ObjectIdentifier(1, 2, 3)
        True

        >>> ObjectIdentifier(1, 2, 4, 5, 6) in ObjectIdentifier(1, 2, 3)
        False
    """
    TAG = 0x06
    value = b''

    @staticmethod
    def decode_large_value(current_char, stream):
        """
        If we encounter a value larger than 127, we have to consume from the
        stram until we encounter a value below 127 and recombine them.

        See: https://msdn.microsoft.com/en-us/library/bb540809(v=vs.85).aspx
        """
        buffer = []
        while current_char > 127:
            buffer.append(current_char ^ 0b10000000)
            current_char = next(stream)
        total = current_char
        for i, digit in enumerate(reversed(buffer)):
            total += digit * 128**(i+1)
        return total

    @staticmethod
    def encode_large_value(value):
        """
        Inverse function of :py:meth:`~.ObjectIdentifier.decode_large_value`
        """
        if value <= 127:
            return [value]
        output = [value & 0b1111111]
        value = value >> 7
        while value:
            output.append(value & 0b1111111 | 0b10000000)
            value = value >> 7
        output.reverse()
        return output

    @classmethod
    def decode(cls, data):
        # Special case for "empty" object identifiers which should be returned
        # as "0"
        if not data:
            return ObjectIdentifier(0)

        # unpack the first byte into first and second sub-identifiers.
        data0 = six.byte2int(data)
        first, second = data0 // 40, data0 % 40
        output = [first, second]

        remaining = six.iterbytes(data[1:])

        for char in remaining:
            # Each node can only contain values from 0-127. Other values need
            # to be combined.
            if char > 127:
                collapsed_value = ObjectIdentifier.decode_large_value(
                    char, remaining)
                output.append(collapsed_value)
                continue
            output.append(char)

        return ObjectIdentifier(*output)

    @staticmethod
    def from_string(value):
        """
        Create an OID from a string
        """

        if not isinstance(value, six.string_types):
            raise TypeError('%r is not of type `str`' % value)

        if value == '.':
            return ObjectIdentifier(1)

        if isinstance(value, str) and value.startswith('.'):
            value = value[1:]

        identifiers = [int(ident, 10) for ident in value.split('.')]
        return ObjectIdentifier(*identifiers)

    def __init__(self, *identifiers):
        # pylint: disable=line-too-long

        # If the user hands in an iterable, instead of positional arguments,
        # make sure we unpack it
        if len(identifiers) == 1 and not isinstance(identifiers[0], int):
            identifiers = [int(ident) for ident in identifiers[0]]

        if len(identifiers) > 1:
            # The first two bytes are collapsed according to X.690
            # See https://en.wikipedia.org/wiki/X.690#BER_encoding
            first, second, rest = identifiers[0], identifiers[1], identifiers[2:]
            first_output = (40*first) + second
        else:
            first_output = identifiers[0]
            rest = tuple()

        # Values above 127 need a special encoding. They get split up into
        # multiple positions.
        exploded_high_values = []
        for char in rest:
            if char > 127:
                exploded_high_values.extend(
                    ObjectIdentifier.encode_large_value(char))
            else:
                exploded_high_values.append(char)

        self.identifiers = tuple(identifiers)
        collapsed_identifiers = [first_output]
        for subidentifier in rest:
            collapsed_identifiers.extend(
                ObjectIdentifier.encode_large_value(subidentifier))
        self.__collapsed_identifiers = tuple(collapsed_identifiers)
        self.length = encode_length(len(self.__collapsed_identifiers))

    def __int__(self):
        if len(self.identifiers) != 1:
            raise ValueError('Only ObjectIdentifier with one node can be '
                             'converted to int. %r is not convertable' % self)
        return self.identifiers[0]

    if six.PY2:  # pragma: no cover
        def __str__(self):
            output = to_bytes([self.TAG])
            if self.__collapsed_identifiers == (0,):
                output += b'\x00'
            else:
                output += self.length + to_bytes(self.__collapsed_identifiers)
            return output

        def __unicode__(self):
            return '.'.join([unicode(_) for _ in self.identifiers])
    else:
        def __str__(self):
            return '.'.join([unicode(_) for _ in self.identifiers])

        def __bytes__(self):
            output = to_bytes([self.TAG])
            if self.__collapsed_identifiers == (0,):
                output += b'\x00'
            else:
                output += self.length + to_bytes(self.__collapsed_identifiers)
            return output

    def __repr__(self):
        return 'ObjectIdentifier(%r)' % (self.identifiers, )

    def __eq__(self, other):
        # pylint: disable=unidiomatic-typecheck, protected-access
        return (type(self) == type(other) and
                self.__collapsed_identifiers == other.__collapsed_identifiers)

    def __len__(self):
        return len(self.identifiers)

    def __contains__(self, other):
        """
        Check if one OID is a child of another.

        TODO: This has been written in the middle of the night! It's messy...
        """
        # pylint: disable=invalid-name

        a, b = other.identifiers, self.identifiers

        # if both have the same amount of identifiers, check for equality
        if len(a) == len(b):
            return a == b

        # if "self" is longer than "other", self cannot be "in" other
        if len(b) > len(a):
            return False

        # For all other cases:
        #   1. zero-fill
        #   2. drop identical items from the front (leaving us with "tail")
        #   3. compare both tails
        zipped = zip_longest(a, b, fillvalue=None)
        tail = []
        for tmp_a, tmp_b in zipped:
            if tmp_a == tmp_b and not tail:
                continue
            tail.append((tmp_a, tmp_b))

        # if we only have Nones in "b", we know that "a" was longer and that it
        # is a direct subtree of "b" (no diverging nodes). Otherwise we would
        # have te divergence in "b", and we can say that "b is contained in a"
        _, unzipped_b = zip(*tail)
        if all([x is None for x in unzipped_b]):
            return True

        # In all other cases we end up with an unmatching tail and know that "b
        # is not contained in a".
        return False

    def __lt__(self, other):
        return self.identifiers < other.identifiers

    def __hash__(self):
        return hash(self.identifiers)

    def __add__(self, other):
        nodes = self.identifiers + other.identifiers
        return ObjectIdentifier(*nodes)

    def __getitem__(self, index):
        return ObjectIdentifier(self.identifiers[index])

    def pythonize(self):
        return '.'.join([unicode(_) for _ in self.identifiers])

    def parentof(self, other):
        """
        Convenience method to check whether this OID is a parent of another OID
        """
        return other in self

    def childof(self, other):
        """
        Convenience method to check whether this OID is a child of another OID
        """
        return self in other


class ObjectDescriptor(Type):
    TAG = 0x07


class External(Type):
    TAG = 0x08


class Real(Type):
    TAG = 0x09
    value = 0.0


class Enumerated(Type):
    TAG = 0x0a


class EmbeddedPdv(Type):
    TAG = 0x0b


class Utf8String(Type):
    TAG = 0x0c
    value = ''


class RelativeOid(Type):
    TAG = 0x0d


class Set(Type):
    TAG = 0x11


class NumericString(Type):
    TAG = 0x12


class PrintableString(Type):
    TAG = 0x13
    value = ''


class T61String(Type):
    TAG = 0x14
    __INITIALISED = False

    def __init__(self, value):
        if not T61String.__INITIALISED:
            t61codec.register()
            self.__INITIALISED = True
        if isinstance(value, six.text_type):
            self.value = value.encode('t61')
        else:
            self.value = value
        self.length = encode_length(len(self.value))

    def __bytes__(self):
        tinfo = TypeInfo(self.TYPECLASS, TypeInfo.PRIMITIVE, self.TAG)
        return to_bytes(tinfo) + self.length + self.value

    def __eq__(self, other):
        # pylint: disable=unidiomatic-typecheck
        return type(self) == type(other) and self.value == other.value

    @classmethod
    def decode(cls, data):
        return cls(data)

    def pythonize(self):
        """
        Convert this object in an appropriate python object
        """
        return self.value.decode('t61')


class VideotexString(Type):
    TAG = 0x15


class IA5String(Type):
    TAG = 0x16


class UtcTime(Type):
    TAG = 0x17


class GeneralizedTime(Type):
    TAG = 0x18


class GraphicString(Type):
    TAG = 0x19


class VisibleString(Type):
    TAG = 0x1a


class GeneralString(Type):
    TAG = 0x1b


class UniversalString(Type):
    TAG = 0x1c


class CharacterString(Type):
    TAG = 0x1d


class BmpString(Type):
    TAG = 0x1e


class EOC(Type):
    TAG = 0x00


class BitString(Type):
    TAG = 0x03



"""
This module contains exceptions for the x.690 protocol
"""


class X690Error(Exception):
    """
    Top-Level exception for everything related to the X690 protocol
    """


class InvalidValueLength(ValueError):
    """
    This error is raised when a value when the length information in the header
    of a "TLV" value does not match the actual value length.

    A likely scenario for this to happen is for example if an IO-operation
    ended prematurely. For example, if the UDP buffer size is too small and the
    remote device returns a packet larger than the buffer.
    """



# pylint: skip-file
# type: ignore


try:
    from unittest.mock import MagicMock
except ImportError:
    from mock import MagicMock  # type: ignore


class AsyncMock(MagicMock):
    async def __call__(self, *args, **kwargs):
        return super(AsyncMock, self).__call__(*args, **kwargs)


class AsyncGenMock(MagicMock):
    async def __call__(self, *args, **kwargs):
        for item in super(AsyncGenMock, self).__call__(*args, **kwargs):
            yield item



"""
This module contains data-types and helpers which are proper to the SNMP
protocol and independent of X.690
"""
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Iterator,
    List,
    Tuple,
    Union,
    cast
)

import six

from .x690.types import (
    Integer,
    Null,
    ObjectIdentifier,
    Sequence,
    Type,
    pop_tlv
)

ERROR_MESSAGES = {
    0: '(noError)',
    1: '(tooBig)',
    2: '(noSuchName)',
    3: '(badValue)',
    4: '(readOnly)',
    5: '(genErr)',
    6: '(noAccess)',
    7: '(wrongType)',
    8: '(wrongLength)',
    9: '(wrongEncoding)',
    10: '(wrongValue)',
    11: '(noCreation)',
    12: '(inconsistentValue)',
    13: '(resourceUnavailable)',
    14: '(commitFailed)',
    15: '(undoFailed)',
    16: '(authorizationError)',
    17: '(notWritable)',
    18: '(inconsistentName)'
}


class VarBind(object):
    '''
    A "VarBind" is a 2-tuple containing an object-identifier and the
    corresponding value.
    '''

    # TODO: This class should be split in two for both the raw and pythonic
    #       API, that would simplify the typing of both "oid" and "value"a lot
    #       and keep things explicit
    oid = ObjectIdentifier(0)  # type: Union[str, ObjectIdentifier]
    value = None  # type: Union[PyType, Type[PyType]]

    def __init__(self, oid, value):
        # type: (Union[ObjectIdentifier, str], PyType) -> None
        if not isinstance(oid, (ObjectIdentifier,) + six.string_types):  # type: ignore
            raise TypeError('OIDs for VarBinds must be ObjectIdentifier or str'
                            ' instances! Your value: %r' % oid)
        if isinstance(oid, six.string_types):
            oid = ObjectIdentifier.from_string(oid)
        self.oid = oid
        self.value = value

    def __iter__(self):
        # type: () -> Iterator[Union[ObjectIdentifier, PyType]]
        return iter([self.oid, self.value])  # type: ignore

    def __getitem__(self, idx):
        return list(self)[idx]

    def __lt__(self, other):
        # type: (Any) -> bool
        return (self.oid, self.value) < (other.oid, other.value)

    def __eq__(self, other):
        # type: (Any) -> bool
        return (self.oid, self.value) == (other.oid, other.value)

    def __hash__(self):
        # type: () -> int
        return hash((self.oid, self.value))

    def __repr__(self):
        # type: () -> str
        return 'VarBind(%r, %r)' % (self.oid, self.value)



from datetime import datetime
from sys import version_info
from typing import Any, Dict, Generic, Iterator, List, Optional, Tuple
from typing import Type as TypeType, TypeVar
from typing import TypeVar, Union

from puresnmp.typevars import TWrappedPyType
from puresnmp.x690.util import TypeInfo


if version_info[0] == 3:
    unicode = str


class Registry(type):
    __registry:  Dict[Tuple[str, int], TypeType[Type[Any]]]
    def __new__(cls, name: str, parents: Tuple[type],
                dict_: Dict[str, Any]) -> 'Registry': ...

    @staticmethod
    def get(typeclass: str, typeid: int) -> TypeType[Type[Any]]: ...


def pop_tlv(data: bytes) -> Tuple[Type[Any], bytes]: ...


# Type --------

class Type(Generic[TWrappedPyType]):
    value: Optional[TWrappedPyType] = None
    @classmethod
    def validate(cls, data: bytes) -> None: ...

    @classmethod
    def from_bytes(cls, data: bytes) -> Type[TWrappedPyType]: ...

    @classmethod
    def decode(cls, data: bytes) -> Type[TWrappedPyType]: ...

    def __bytes__(self) -> bytes: ...
    def __repr__(self) -> str: ...
    def pythonize(self) -> Any: ...
    def pretty(self) -> str: ...

    if version_info[0] == 2:
        def __unicode__(self) -> unicode: ...
        def __str__(self) -> str: ...


class UnknownType(Type[bytes]):
    value = b''

    def __init__(self, tag: int, value: bytes,
                 typeinfo: Optional[TypeInfo] = None) -> None: ...

    @classmethod
    def from_bytes(cls, data: bytes) -> 'UnknownType': ...


class NonASN1Type(UnknownType):
    def __init__(self, tag: int, value: Any) -> None: ...


class Boolean(Type[bool]):
    ...


class Null(Type[None]):
    ...


class OctetString(Type[bytes]):
    def __init__(self, value: Union[unicode, bytes]) -> None: ...


class Sequence(Type[List[Type[Any]]]):  # type: ignore
    def __init__(self, *items: Type[Any]) -> None: ...


class Integer(Type[int]):
    def __init__(self, value: int) -> None: ...


class ObjectIdentifier(Type[str]):
    def __init__(self, *identifiers: int) -> None: ...
    def __contains__(self, other: 'ObjectIdentifier') -> bool: ...

    @staticmethod
    def decode_large_value(
        current_char: int, stream: Iterator[int]) -> int: ...

    @staticmethod
    def encode_large_value(value: int) -> List[int]: ...
    @staticmethod
    def from_string(value: str) -> 'ObjectIdentifier': ...
    def pythonize(self) -> str: ...
    def parentof(self, other: 'ObjectIdentifier') -> bool: ...
    def childof(self, other: 'ObjectIdentifier') -> bool: ...


class ObjectDescriptor(Type[str]):
    ...


class External(Type[bytes]):
    ...


class Real(Type[float]):
    ...


class Enumerated(Type[List[Any]]):  # type: ignore
    ...


class EmbeddedPdv(Type[bytes]):
    ...


class Utf8String(Type[str]):
    ...


class RelativeOid(Type[str]):
    ...


class Set(Type[bytes]):
    ...


class NumericString(Type[str]):
    ...


class PrintableString(Type[str]):
    ...


class T61String(Type[str]):
    ...


class VideotexString(Type[str]):
    ...


class IA5String(Type[str]):
    ...


class UtcTime(Type[datetime]):
    ...


class GeneralizedTime(Type[datetime]):
    ...


class GraphicString(Type[str]):
    ...


class VisibleString(Type[str]):
    ...


class GeneralString(Type[str]):
    ...


class UniversalString(Type[str]):
    ...


class CharacterString(Type[str]):
    ...


class BmpString(Type[str]):
    ...


class EOC(Type[bytes]):
    ...


class BitString(Type[str]):
    ...



"""
Utility functions for working with the X.690 and related standards.
"""
from __future__ import division, print_function, unicode_literals

from binascii import hexlify, unhexlify
from collections import namedtuple
from typing import TYPE_CHECKING

import six

from ..const import Length

if TYPE_CHECKING:  # pragma: no cover
    # pylint: disable=unused-import, cyclic-import
    from typing import Any, Dict, Iterable, List, Union, Tuple
    from .types import Type
    from puresnmp.typevars import PyType

if six.PY2:  # pragma: no cover

    def to_bytes(obj):
        # type: (Any) -> bytes
        """
        Converts an object to bytes.

        If it has a ``__bytes__`` attribute, uses this for conversion,
        otherwise converts first to ``bytearray``. This was only implemented
        for Python 2/3 consistency
        """
        if hasattr(obj, '__bytes__'):
            return bytes(obj)
        else:
            return bytes(bytearray(obj))

    def int_from_bytes(bytes_, byteorder, signed=False):
        # type: (bytes, str, bool) -> int
        """
        Converts a number to bytes
        """
        bytes_ = bytearray(bytes_)
        if byteorder == 'little':
            little_ordered = list(bytes_)
        elif byteorder == 'big':
            little_ordered = list(reversed(bytes_))
        n = sum(b << 8*i for i, b in enumerate(little_ordered))
        if signed and little_ordered and (little_ordered[-1] & 0x80):
            n -= 1 << 8*len(little_ordered)
        return n

else:
    unicode = str  # pylint: disable=invalid-name
    int_from_bytes = int.from_bytes  # pylint: disable=invalid-name

    def to_bytes(obj):
        # type: (Any) -> bytes
        """
        Converts an instance to bytes, and if it does not work, adds helpful
        details to the raised ``TypeError``
        """
        try:
            return bytes(obj)
        except TypeError as exc:
            raise TypeError(exc.args[0] + ' on type {}'.format(type(obj)))

LengthValue = namedtuple('LengthValue', 'length value')


class TypeInfo(namedtuple('TypeInfo', 'cls priv_const tag')):
    """
    Decoded structure for an X.690 "type" octet. Example::

        >>> TypeInfo.from_bytes(b'\\x30')
        TypeInfo(cls='universal', priv_const='constructed', tag=16)

    The structure contains 3 fields:

    cls
        The typeclass (either :py:attr:`~.UNIVERSAL`, :py:attr:`~.APPLICATION`,
        :py:attr:`~.CONTEXT` or :py:attr:`~.CONSTRUCTED`)

    priv_const
        Whether the value is :py:attr:`~.CONSTRUCTED` or :py:attr:`~.PRIMITIVE`

    tag
        The actual type identifier.

    The instance also keeps the raw value as it was seen in the ``_raw_value``
    attribute.
    """

    UNIVERSAL = 'universal'
    APPLICATION = 'application'
    CONTEXT = 'context'
    PRIVATE = 'private'
    PRIMITIVE = 'primitive'
    CONSTRUCTED = 'constructed'
    _raw_value = None

    @staticmethod
    def from_bytes(data):
        # type: ( Union[int, bytes] ) -> TypeInfo
        """
        Given one octet, extract the separate fields and return a TypeInfo
        instance::

            >>> TypeInfo.from_bytes(b'\\x30')
            TypeInfo(cls='universal', priv_const='constructed', tag=16)
        """
        # pylint: disable=attribute-defined-outside-init

        if isinstance(data, (bytes, bytearray)):
            data = int_from_bytes(data, 'big')
        # pylint: disable=protected-access
        if data == 0b11111111:
            raise NotImplementedError('Long identifier types are not yet '
                                      'implemented')
        cls_hint = (data & 0b11000000) >> 6
        pc_hint = (data & 0b00100000) >> 5
        value = data & 0b00011111

        if cls_hint == 0b00:
            cls = TypeInfo.UNIVERSAL
        elif cls_hint == 0b01:
            cls = TypeInfo.APPLICATION
        elif cls_hint == 0b10:
            cls = TypeInfo.CONTEXT
        elif cls_hint == 0b11:
            cls = TypeInfo.PRIVATE
        else:
            pass  # Impossible case (2 bits can only have 4 combinations).

        priv_const = TypeInfo.CONSTRUCTED if pc_hint else TypeInfo.PRIMITIVE

        instance = TypeInfo(cls, priv_const, value)
        instance._raw_value = data
        return instance

    def __bytes__(self):
        # type: () -> bytes
        # pylint: disable=invalid-name
        if self.cls == TypeInfo.UNIVERSAL:
            cls = 0b00
        elif self.cls == TypeInfo.APPLICATION:
            cls = 0b01
        elif self.cls == TypeInfo.CONTEXT:
            cls = 0b10
        elif self.cls == TypeInfo.PRIVATE:
            cls = 0b11
        else:
            raise ValueError('Unexpected class for type info')

        if self.priv_const == TypeInfo.CONSTRUCTED:
            priv_const = 0b01
        elif self.priv_const == TypeInfo.PRIMITIVE:
            priv_const = 0b00
        else:
            raise ValueError('Unexpected primitive/constructed for type info')

        output = cls << 6 | priv_const << 5 | self.tag
        return to_bytes([output])

    if six.PY2:  # pragma: no cover
        def __unicode__(self):
            return repr(self)

        def __str__(self):
            return self.__bytes__()


def encode_length(value):
    # type: (int) -> bytes
    """
    This function encodes the length of a variable into bytes conforming to the
    rules defined in :term:`X.690`: The "length" field must be specially
    encoded for values above 127.  Additionally, from :term:`X.690`:

        8.1.3.2 A sender shall:

            a) use the definite form (see 8.1.3.3) if the encoding is
               primitive;
            b) use either the definite form (see 8.1.3.3) or the indefinite
               form (see 8.1.3.6), a sender's option, if the encoding is
               constructed and all immediately available;
            c) use the indefinite form (see 8.1.3.6) if the encoding is
               constructed and is not all immediately available.

    See also: https://en.wikipedia.org/wiki/X.690#Length_octets

    Example::

        >>> encode_length(16)    # no need for special encoding.
        b'\\x10'
        >>> encode_length(200)   # > 127, needs to be specially encoded.
        b'\\x81\\xc8'
    """
    if value == Length.INDEFINITE:
        return to_bytes([0b10000000])

    if value < 127:
        return to_bytes([value])

    output = []  # type: List[int]
    while value > 0:
        value, remainder = value // 256, value % 256
        output.insert(0, remainder)

    # prefix length information
    output = [0b10000000 | len(output)] + output
    return to_bytes(output)


def decode_length(data):
    # type: ( bytes ) -> LengthValue
    """
    Given a bytes object, which starts with the length information of a TLV
    value, returns a namedtuple with the length and the remaining bytes. So,
    given a TLV value, this function takes the "LV" part as input, parses the
    length information and returns the length plus the remaining "V" part
    (including any subsequent bytes).

    For values which are longer than 127 bytes, the length must be encoded into
    an unknown amount of "length" bytes. This function reads as many bytes as
    needed for the length. The return value contains the parsed length in
    number of bytes, and the remaining data bytes which follow the length
    bytes.

    Examples::

        >>> # length > 127, consume multiple length bytes
        >>> decode_length(b'\\x81\\xc8...')
        LengthValue(length=200, value=b'...')

        >>> # length <= 127, consume one length byte
        >>> decode_length(b'\\x10...')
        LengthValue(length=16, value=b'...')

    TODO: Upon rereading this, I wonder if it would not make more sense to take
          the complete TLV content as input.
    """
    data0 = six.byte2int(data)
    if data0 == 0b11111111:
        # reserved
        raise NotImplementedError('This is a reserved case in X690')

    if data0 & 0b10000000 == 0:
        # definite short form
        output = int_from_bytes([data0], 'big')
        data = data[1:]
    elif data0 ^ 0b10000000 == 0:
        # indefinite form
        raise NotImplementedError('Indefinite lenghts are '
                                  'not yet implemented!')
    else:
        # definite long form
        num_octets = int_from_bytes([data0 ^ 0b10000000], 'big')
        value_octets = data[1:1+num_octets]
        output = int_from_bytes(value_octets, 'big')
        data = data[num_octets + 1:]
    return LengthValue(output, data)


def visible_octets(data):
    # type: ( bytes ) -> str
    """
    Returns a geek-friendly (hexdump)  output of a bytes object.

    Developer note:
        This is not super performant. But it's not something that's supposed to
        be run during normal operations (mostly for testing and debugging).  So
        performance should not be an issue, and this is less obfuscated than
        existing solutions.

    Example::

        >>> from os import urandom
        >>> print(visible_octets(urandom(40)))
        99 1f 56 a9 25 50 f7 9b  95 7e ff 80 16 14 88 c5   ..V.%P...~......
        f3 b4 83 d4 89 b2 34 b4  71 4e 5a 69 aa 9f 1d f8   ......4.qNZi....
        1d 33 f9 8e f1 b9 12 e9                            .3......

    """
    hexed = hexlify(data).decode('ascii')
    tuples = [''.join((a, b)) for a, b in zip(hexed[::2], hexed[1::2])]
    line = []
    output = []
    ascii_column = []
    for idx, octet in enumerate(tuples):
        line.append(octet)
        # only use printable characters in ascii output
        ascii_column.append(octet if 32 <= int(octet, 16) < 127 else '2e')
        if (idx+1) % 8 == 0:
            line.append('')
        if (idx+1) % 8 == 0 and (idx+1) % 16 == 0:
            raw_ascii = unhexlify(''.join(ascii_column))
            raw_ascii = raw_ascii.replace(b'\\n z', b'.')
            ascii_column = []
            output.append('%-50s %s' % (' '.join(line),
                                        raw_ascii.decode('ascii')))
            line = []
    raw_ascii = unhexlify(''.join(ascii_column))
    raw_ascii = raw_ascii.replace(b'\\n z', b'.')
    output.append('%-50s %s' % (' '.join(line), raw_ascii.decode('ascii')))
    line = []
    return '\n'.join(output)


def tablify(varbinds, num_base_nodes=0, base_oid=''):
    # type: ( Iterable[Tuple[Any, Any]], int, str ) -> List[Dict[str, Any]]
    """
    Converts a list of varbinds into a table-like structure. *num_base_nodes*
    can be used for table which row-ids consist of multiple OID tree nodes. By
    default, the last node is considered the row ID, and the second-last is the
    column ID. Example:

    By default, for the table-cell at OID ``1.2.3.4.5``, ``4`` is the column
    index and ``5`` is the row index.

    Using ``num_base_nodes=2`` will only use the first two nodes (``1.2``) as
    table-identifier, so ``3`` becomes the column index, and ``4.5`` becomes
    the row index.

    The output should *not* be considered ordered in any way. If you need it
    sorted, you must sort it after retrieving the table from this function!

    Each element of the output is a dictionary where each key is the column
    index. By default the index ``0`` will be added, representing the row ID.

    Example::

        >>> data = [
        >>>     (ObjectIdentifier.from_string('1.2.1.1'), 'row 1 col 1'),
        >>>     (ObjectIdentifier.from_string('1.2.1.2'), 'row 2 col 1'),
        >>>     (ObjectIdentifier.from_string('1.2.2.1'), 'row 1 col 2'),
        >>>     (ObjectIdentifier.from_string('1.2.2.2'), 'row 2 col 2'),
        >>> ]
        >>> tablify(data)
        [
            {'0': '1', '1': 'row 1 col 1', '2': 'row 1 col 2'},
            {'0': '2', '1': 'row 2 col 1', '2': 'row 2 col 2'},
        ]


    Example with longer row ids (using the *first* two as table identifiers)::

        >>> data = [
        >>>     (ObjectIdentifier.from_string('1.2.1.5.10'), 'row 5.10 col 1'),
        >>>     (ObjectIdentifier.from_string('1.2.1.6.10'), 'row 6.10 col 1'),
        >>>     (ObjectIdentifier.from_string('1.2.2.5.10'), 'row 5.10 col 2'),
        >>>     (ObjectIdentifier.from_string('1.2.2.6.10'), 'row 6.10 col 2'),
        >>> ]
        >>> tablify(data, num_base_nodes=2)
        [
            {'0': '5.10', '1': 'row 5.10 col 1', '2': 'row 5.10 col 2'},
            {'0': '6.10', '1': 'row 6.10 col 1', '2': 'row 6.10 col 2'},
        ]
    """
    if isinstance(base_oid, str) and base_oid:
        # This import needs to be delayed to avoid circular imports
        from puresnmp.x690.types import ObjectIdentifier
        base_oid_parsed = ObjectIdentifier.from_string(base_oid)
        # Each table has a sub-index for the table "entry" so the number of
        # base-nodes needs to be incremented by 1
        num_base_nodes = len(base_oid_parsed)  #  type: ignore

    rows = {}  # type: Dict[str, Dict[str, Type[PyType]]]
    for oid, value in varbinds:
        if num_base_nodes:
            tail = oid.identifiers[num_base_nodes:]
            col_id, row_id = tail[0], tail[1:]
            row_id = '.'.join([unicode(node) for node in row_id])
        else:
            col_id = unicode(oid.identifiers[-2])
            row_id = unicode(oid.identifiers[-1])
        tmp = {
            '0': row_id,
        }
        row = rows.setdefault(row_id, tmp)
        row[unicode(col_id)] = value
    return list(rows.values())
